import math
import torch

import torch.nn.functional as F

from torch import Tensor
from typing import Any
from typing import Dict
from typing import Type
from typing import Tuple
from typing import Callable
from typing import Iterable
from typing import Optional
from torch.optim.optimizer import Optimizer


optimizer_dict: Dict[str, Type[Optimizer]] = {}


def register_optimizer(name: str) -> Callable:
    def _register(cls_: Type) -> Type:
        global optimizer_dict
        optimizer_dict[name] = cls_
        return cls_

    return _register


register_optimizer("sgd")(torch.optim.SGD)
register_optimizer("adam")(torch.optim.Adam)
register_optimizer("adamw")(torch.optim.AdamW)
register_optimizer("rmsprop")(torch.optim.RMSprop)


# AdamP
# reference : https://github.com/clovaai/AdamP


def channel_view(net: Tensor) -> Tensor:
    return net.view(net.shape[0], -1)


def layer_view(net: Tensor) -> Tensor:
    return net.view(1, -1)


def cosine_similarity(x: Tensor, y: Tensor, eps: float, view_func: Callable) -> Tensor:
    x = view_func(x)
    y = view_func(y)
    return F.cosine_similarity(x, y, dim=1, eps=eps).abs_()


def projection(
    p: Tensor,
    grad: Tensor,
    perturb: Tensor,
    delta: float,
    wd_ratio: float,
    eps: float,
) -> Tuple[Tensor, float]:
    wd = 1.0
    expand_size = [-1] + [1] * (len(p.shape) - 1)
    for view_func in [channel_view, layer_view]:
        cosine_sim = cosine_similarity(grad, p.data, eps, view_func)
        if cosine_sim.max() < delta / math.sqrt(view_func(p.data).shape[1]):
            p_n = p.data / view_func(p.data).norm(dim=1).view(expand_size).add_(eps)
            perturb -= p_n * view_func(p_n * perturb).sum(dim=1).view(expand_size)
            wd = wd_ratio
            return perturb, wd
    return perturb, wd


@register_optimizer("adamp")
class AdamP(Optimizer):
    def __init__(
        self,
        params: Iterable[Tensor],
        lr: float = 1.0e-3,
        betas: Tuple[float, float] = (0.9, 0.999),
        eps: float = 1.0e-8,
        weight_decay: float = 0.0,
        delta: float = 0.1,
        wd_ratio: float = 0.1,
        nesterov: bool = False,
    ):
        defaults = dict(
            lr=lr,
            betas=betas,
            eps=eps,
            weight_decay=weight_decay,
            delta=delta,
            wd_ratio=wd_ratio,
            nesterov=nesterov,
        )
        super(AdamP, self).__init__(params, defaults)

    def step(self, closure: Optional[Callable] = None) -> Any:
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue

                grad = p.grad.data
                beta1, beta2 = group["betas"]
                nesterov = group["nesterov"]

                state = self.state[p]
                if len(state) == 0:
                    state["step"] = 0
                    state["exp_avg"] = torch.zeros_like(p.data)
                    state["exp_avg_sq"] = torch.zeros_like(p.data)

                # Adam
                exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]

                state["step"] += 1
                bias_correction1 = 1.0 - beta1 ** state["step"]
                bias_correction2 = 1.0 - beta2 ** state["step"]
                bias_correction2 = math.sqrt(bias_correction2)

                exp_avg.mul_(beta1).add_(grad, alpha=1.0 - beta1)
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1.0 - beta2)

                denom = (exp_avg_sq.sqrt() / bias_correction2).add_(group["eps"])
                step_size = group["lr"] / bias_correction1

                if nesterov:
                    perturb = (beta1 * exp_avg + (1 - beta1) * grad) / denom
                else:
                    perturb = exp_avg / denom

                # Projection
                wd_ratio = 1.0
                if len(p.shape) > 1:
                    perturb, wd_ratio = projection(
                        p,
                        grad,
                        perturb,
                        group["delta"],
                        group["wd_ratio"],
                        group["eps"],
                    )

                # Weight decay
                if group["weight_decay"] > 0.0:
                    p.data.mul_(1.0 - group["lr"] * group["weight_decay"] * wd_ratio)

                # Step
                p.data.add_(perturb, alpha=-step_size)

        return loss


__all__ = ["optimizer_dict", "register_optimizer"]



import os

from typing import Any
from typing import Dict
from cftool.misc import OPTBase


class OPTClass(OPTBase):
    cache_dir: str
    meta_settings: Dict[str, Any]

    @property
    def env_key(self) -> str:
        return "CFLEARN_ENV"

    @property
    def defaults(self) -> Dict[str, Any]:
        return dict(
            cache_dir=os.path.join(os.path.expanduser("~"), ".cache", "carefree-learn"),
            meta_settings={},
        )

    @property
    def data_cache_dir(self) -> str:
        return os.path.join(self.cache_dir, "data")


OPT = OPTClass()


__all__ = [
    "OPT",
]



# based on https://github.com/isl-org/MiDaS

import math
import torch

import numpy as np
import torch.nn as nn

from torchvision.transforms import Compose

from .core.dpt_depth import DPTDepthModel
from .core.midas_net import MidasNet
from .core.midas_net_custom import MidasNet_small
from .core.transforms import Resize, NormalizeImage, PrepareForNet
from .....misc.toolkit import download_model

try:
    import cv2
except:
    cv2 = None
try:
    import timm
except:
    timm = None


ISL_TAGS = {
    "dpt_large": "dpt_large-midas-2f21e586",
    "dpt_hybrid": "dpt_hybrid-midas-501f0c75",
    "midas_v21": "",
    "midas_v21_small": "",
}


def disabled_train(self, mode=True):
    """Overwrite model.train with this function to make sure train/eval mode
    does not change anymore."""
    return self


def load_midas_transform(model_type):
    # https://github.com/isl-org/MiDaS/blob/master/run.py
    # load transform only
    if model_type == "dpt_large":  # DPT-Large
        net_w, net_h = 384, 384
        resize_mode = "minimal"
        normalization = NormalizeImage(mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5])

    elif model_type == "dpt_hybrid":  # DPT-Hybrid
        net_w, net_h = 384, 384
        resize_mode = "minimal"
        normalization = NormalizeImage(mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5])

    elif model_type == "midas_v21":
        net_w, net_h = 384, 384
        resize_mode = "upper_bound"
        normalization = NormalizeImage(
            mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
        )

    elif model_type == "midas_v21_small":
        net_w, net_h = 256, 256
        resize_mode = "upper_bound"
        normalization = NormalizeImage(
            mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
        )

    else:
        assert (
            False
        ), f"model_type '{model_type}' not implemented, use: --model_type large"

    transform = Compose(
        [
            Resize(
                net_w,
                net_h,
                resize_target=None,
                keep_aspect_ratio=True,
                ensure_multiple_of=32,
                resize_method=resize_mode,
                image_interpolation_method=cv2.INTER_CUBIC,
            ),
            normalization,
            PrepareForNet(),
        ]
    )

    return transform


def load_model(model_type):
    # https://github.com/isl-org/MiDaS/blob/master/run.py
    # load network
    model_tag = ISL_TAGS[model_type]
    model_path = download_model(model_tag)
    if model_type == "dpt_large":  # DPT-Large
        model = DPTDepthModel(
            path=model_path,
            backbone="vitl16_384",
            non_negative=True,
        )
        net_w, net_h = 384, 384
        resize_mode = "minimal"
        normalization = NormalizeImage(mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5])

    elif model_type == "dpt_hybrid":  # DPT-Hybrid
        model = DPTDepthModel(
            path=model_path,
            backbone="vitb_rn50_384",
            non_negative=True,
        )
        net_w, net_h = 384, 384
        resize_mode = "minimal"
        normalization = NormalizeImage(mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5])

    elif model_type == "midas_v21":
        model = MidasNet(model_path, non_negative=True)
        net_w, net_h = 384, 384
        resize_mode = "upper_bound"
        normalization = NormalizeImage(
            mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
        )

    elif model_type == "midas_v21_small":
        model = MidasNet_small(
            model_path,
            features=64,
            backbone="efficientnet_lite3",
            exportable=True,
            non_negative=True,
            blocks={"expand": True},
        )
        net_w, net_h = 256, 256
        resize_mode = "upper_bound"
        normalization = NormalizeImage(
            mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
        )

    else:
        print(f"model_type '{model_type}' not implemented, use: --model_type large")
        assert False

    transform = Compose(
        [
            Resize(
                net_w,
                net_h,
                resize_target=None,
                keep_aspect_ratio=True,
                ensure_multiple_of=32,
                resize_method=resize_mode,
                image_interpolation_method=cv2.INTER_CUBIC,
            ),
            normalization,
            PrepareForNet(),
        ]
    )

    return model.eval(), transform


class MiDaSInference(nn.Module):
    MODEL_TYPES_TORCH_HUB = ["DPT_Large", "DPT_Hybrid", "MiDaS_small"]
    MODEL_TYPES_ISL = [
        "dpt_large",
        "dpt_hybrid",
        "midas_v21",
        "midas_v21_small",
    ]

    def __init__(self, model_type):
        super().__init__()
        assert model_type in self.MODEL_TYPES_ISL
        model, _ = load_model(model_type)
        self.model = model
        self.model.train = disabled_train

    def forward(self, x):
        with torch.no_grad():
            prediction = self.model(x)
        return prediction


class MiDaSAPI:
    def __init__(self, device: torch.device):
        if cv2 is None:
            raise ValueError("`cv2` is needed for `MiDaSAPI`")
        if timm is None:
            raise ValueError("`timm` is needed for `MiDaSAPI`")
        self.model = MiDaSInference(model_type="dpt_hybrid").to(device)
        self.device = device

    @property
    def dtype(self) -> torch.dtype:
        return list(self.model.parameters())[0].dtype

    def to(self, device: torch.device, *, use_half: bool) -> None:
        if use_half:
            self.model.half()
        else:
            self.model.float()
        self.device = device
        self.model.to(device)

    def to_self(self, net: torch.Tensor) -> torch.Tensor:
        return net.to(self.dtype).to(self.device)

    def detect_depth(self, input_image: np.ndarray) -> np.ndarray:
        assert input_image.ndim == 3
        image_depth = input_image
        with torch.no_grad():
            image_depth = self.to_self(torch.from_numpy(image_depth))
            image_depth = image_depth / 127.5 - 1.0
            image_depth = image_depth[None].permute(0, 3, 1, 2).contiguous()
            depth = self.model(image_depth)[0]

            depth_pt = depth.clone()
            depth_pt -= torch.min(depth_pt)
            depth_pt /= torch.max(depth_pt)
            depth_image = depth_pt.cpu().numpy()
            depth_image = (depth_image * 255.0).clip(0, 255).astype(np.uint8)

            return depth_image

    def detect_normal(
        self,
        input_image: np.ndarray,
        a: float = math.pi * 2.0,
        bg_th: float = 0.1,
    ) -> np.ndarray:
        assert input_image.ndim == 3
        image_depth = input_image
        with torch.no_grad():
            image_depth = self.to_self(torch.from_numpy(image_depth))
            image_depth = image_depth / 127.5 - 1.0
            image_depth = image_depth[None].permute(0, 3, 1, 2).contiguous()
            depth = self.model(image_depth)[0]

            depth_pt = depth.clone()
            depth_pt -= torch.min(depth_pt)
            depth_pt /= torch.max(depth_pt)
            depth_pt = depth_pt.cpu().numpy()

            depth_np = depth.cpu().numpy()
            x = cv2.Sobel(depth_np, cv2.CV_32F, 1, 0, ksize=3)
            y = cv2.Sobel(depth_np, cv2.CV_32F, 0, 1, ksize=3)
            z = np.ones_like(x) * a
            x[depth_pt < bg_th] = 0
            y[depth_pt < bg_th] = 0
            normal = np.stack([x, y, z], axis=2)
            normal /= np.sum(normal**2.0, axis=2, keepdims=True) ** 0.5
            normal_image = (normal * 127.5 + 127.5).clip(0, 255).astype(np.uint8)

            return normal_image


__all__ = [
    "MiDaSAPI",
]



import numpy as np

from enum import Enum
from typing import Any
from typing import Dict
from typing import Optional
from dataclasses import dataclass

from .file import TArrayPair
from .file import FileParserBlock
from ....schema import DataBundle
from ....schema import INoInitDataBlock


class NanReplaceMethod(str, Enum):
    MEAN = "mean"
    MEDIAN = "median"


class NanDropStrategy(str, Enum):
    NONE = "none"
    DROP_Y = "drop_y"
    DROP_ALL = "drop_all"


@dataclass
class MLNanHandlerConfig:
    drop_strategy: NanDropStrategy = NanDropStrategy.DROP_Y
    replace_method: NanReplaceMethod = NanReplaceMethod.MEDIAN


def drop_nan(
    x: np.ndarray,
    y: Optional[np.ndarray],
    x_nan_mask: np.ndarray,
    y_nan_mask: Optional[np.ndarray],
    *,
    only_drop_y: bool,
) -> TArrayPair:
    if y_nan_mask is not None:
        y_nan_mask = np.any(y_nan_mask, axis=1)
    if not only_drop_y:
        x_nan_mask = np.any(x_nan_mask, axis=1)
        if y_nan_mask is None:
            nan_mask = x_nan_mask
        else:
            nan_mask = x_nan_mask | y_nan_mask
    else:
        if y_nan_mask is None:
            return x, y
        nan_mask = y_nan_mask
    valid_mask = ~nan_mask
    return x[valid_mask], None if y is None else y[valid_mask]


def replace_nan(
    data: np.ndarray,
    nan_mask: np.ndarray,
    idx: int,
    column_name: str,
    replacement: Optional[float],
) -> None:
    array = data[..., idx]
    nan_mask = nan_mask[..., idx]
    if not np.any(nan_mask):
        return
    if replacement is None:
        msg = f"nan values occurred at column {column_name} but replacement is not available"
        raise ValueError(msg)
    array[nan_mask] = replacement
    data[..., idx] = array


@INoInitDataBlock.register("ml_nan_handler")
class NanHandlerBlock(INoInitDataBlock):
    config: MLNanHandlerConfig  # type: ignore
    drop_strategy: NanDropStrategy
    has_nan: Dict[str, bool]
    replacements: Dict[str, float]
    label_has_nan: Dict[str, bool]
    label_replacements: Dict[str, float]

    # inheritance

    def to_info(self) -> Dict[str, Any]:
        return dict(
            drop_strategy=self.drop_strategy,
            has_nan=self.has_nan,
            replacements=self.replacements,
            label_has_nan=self.label_has_nan,
            label_replacements=self.label_replacements,
        )

    def transform(self, bundle: DataBundle, for_inference: bool) -> DataBundle:
        return self._transform(bundle, for_inference)

    def fit_transform(self, bundle: DataBundle) -> DataBundle:
        x_tr = bundle.x_train
        y_tr = bundle.y_train
        if not isinstance(x_tr, np.ndarray):
            raise ValueError("`NanHandlerBlock` should be used on numpy features.")
        if not isinstance(y_tr, np.ndarray):
            raise ValueError("`NanHandlerBlock` should be used on numpy labels.")
        x_tr_nan_mask = np.isnan(x_tr)
        ytr_nan_mask = np.isnan(y_tr)
        self.drop_strategy = self.config.drop_strategy
        self.has_nan = {}
        self.replacements = {}
        self.label_has_nan = {}
        self.label_replacements = {}
        self._fit(self.has_nan, self.replacements, x_tr, x_tr_nan_mask)
        self._fit(self.label_has_nan, self.label_replacements, y_tr, ytr_nan_mask)
        return self._transform(
            bundle,
            False,
            x_tr_nan_mask,
            ytr_nan_mask,
            None if bundle.x_valid is None else np.isnan(bundle.x_valid),
            None if bundle.y_valid is None else np.isnan(bundle.y_valid),
        )

    # api

    @property
    def file_parser(self) -> Optional[FileParserBlock]:
        return self.try_get_previous(FileParserBlock)

    # internal

    def _fit(
        self,
        has_nan: Dict[str, bool],
        replacements: Dict[str, float],
        data: np.ndarray,
        data_nan_mask: np.ndarray,
    ) -> None:
        for idx in range(data.shape[-1]):
            str_idx = str(idx)
            array = data[..., idx]
            not_nan = array[~data_nan_mask[..., idx]]
            has_nan[str_idx] = len(array) != len(not_nan)
            if len(not_nan) == 0:
                replacements[str_idx] = 0.0
            else:
                replacements[str_idx] = getattr(np, self.config.replace_method)(not_nan)

    def _get_column_name(self, idx: int, is_label: bool) -> str:
        file_parser = self.file_parser
        if file_parser is None:
            return str(idx)
        if is_label:
            return f"'{file_parser.label_header[idx]}'"
        return f"'{file_parser.feature_header[idx]}'"

    def _transform(
        self,
        bundle: DataBundle,
        for_inference: bool,
        x_train_nan_mask: Optional[np.ndarray] = None,
        y_train_nan_mask: Optional[np.ndarray] = None,
        x_valid_nan_mask: Optional[np.ndarray] = None,
        y_valid_nan_mask: Optional[np.ndarray] = None,
    ) -> DataBundle:
        if not isinstance(bundle.x_train, np.ndarray):
            raise ValueError("`NanHandlerBlock` should be used on numpy features.")
        if bundle.y_train is not None and not isinstance(bundle.y_train, np.ndarray):
            raise ValueError("`NanHandlerBlock` should be used on numpy labels.")
        if x_train_nan_mask is None:
            x_train_nan_mask = np.isnan(bundle.x_train)
        if y_train_nan_mask is None and bundle.y_train is not None:
            y_train_nan_mask = np.isnan(bundle.y_train)
        if x_valid_nan_mask is None and bundle.x_valid is not None:
            x_valid_nan_mask = np.isnan(bundle.x_valid)
        if y_valid_nan_mask is None and bundle.y_valid is not None:
            y_valid_nan_mask = np.isnan(bundle.y_valid)
        # drop nan
        if not for_inference and self.drop_strategy != NanDropStrategy.NONE:
            only_drop_y = self.drop_strategy == NanDropStrategy.DROP_Y
            bundle.x_train, bundle.y_train = drop_nan(
                bundle.x_train,
                bundle.y_train,
                x_train_nan_mask,
                y_train_nan_mask,
                only_drop_y=only_drop_y,
            )
            if bundle.x_valid is not None:
                bundle.x_valid, bundle.y_valid = drop_nan(
                    bundle.x_valid,
                    bundle.y_valid,
                    x_valid_nan_mask,
                    y_valid_nan_mask,
                    only_drop_y=only_drop_y,
                )
        # replace nan
        for idx in range(bundle.x_train.shape[-1]):
            cn = self._get_column_name(idx, False)
            replacement = self.replacements.get(str(idx))
            replace_nan(bundle.x_train, x_train_nan_mask, idx, cn, replacement)
            if bundle.x_valid is not None:
                replace_nan(bundle.x_valid, x_valid_nan_mask, idx, cn, replacement)
        if bundle.y_train is not None:
            for idx in range(bundle.y_train.shape[-1]):
                cn = self._get_column_name(idx, True)
                replacement = self.label_replacements.get(str(idx))
                replace_nan(bundle.y_train, y_train_nan_mask, idx, cn, replacement)
                if bundle.y_valid is not None:
                    replace_nan(bundle.y_valid, y_valid_nan_mask, idx, cn, replacement)
        return bundle


__all__ = [
    "NanReplaceMethod",
    "NanDropStrategy",
    "MLNanHandlerConfig",
    "NanHandlerBlock",
]



import torch

from typing import Dict
from typing import Optional

from .base import MLModel
from ...schema import MLEncoderSettings
from ...schema import MLGlobalEncoderSettings


@MLModel.register("linear")
class Linear(MLModel):
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        num_history: int = 1,
        *,
        bias: bool = True,
        encoder_settings: Optional[Dict[str, MLEncoderSettings]] = None,
        global_encoder_settings: Optional[MLGlobalEncoderSettings] = None,
    ):
        super().__init__(
            encoder_settings=encoder_settings,
            global_encoder_settings=global_encoder_settings,
        )
        if self.encoder is not None:
            input_dim += self.encoder.dim_increment
        input_dim *= num_history
        self.net = torch.nn.Linear(input_dim, output_dim, bias)

    def forward(self, net: torch.Tensor) -> torch.Tensor:
        if len(net.shape) > 2:
            net = net.contiguous().view(len(net), -1)
        return self.net(net)


__all__ = ["Linear"]



import os
import torch

import numpy as np

from .utils import pred_lines
from .models.mbv2_mlsd_large import MobileV2_MLSD_Large
from .....misc.toolkit import download_model

try:
    import cv2
except:
    cv2 = None


class MLSDDetector:
    def __init__(self, device: torch.device):
        if cv2 is None:
            raise ValueError("`cv2` is needed for `MLSDdetector`")
        model_path = download_model("mlsd_large_512_fp32")
        model = MobileV2_MLSD_Large()
        model.load_state_dict(torch.load(model_path), strict=True)
        self.model = model.to(device).eval()
        self.device = device
        self.use_half = False

    def to(self, device: torch.device, *, use_half: bool) -> None:
        if use_half:
            self.model.half()
        else:
            self.model.float()
        self.model.to(device)
        self.device = device
        self.use_half = use_half

    def __call__(self, input_image, value_threshold: float, distance_threshold: float):
        def to(tensor):
            tensor = tensor.to(self.device)
            tensor = tensor.half() if self.use_half else tensor.float()
            return tensor

        assert input_image.ndim == 3
        img = input_image
        img_output = np.zeros_like(img)
        try:
            with torch.no_grad():
                lines = pred_lines(
                    img,
                    self.model,
                    to,
                    [img.shape[0], img.shape[1]],
                    value_threshold,
                    distance_threshold,
                )
                for line in lines:
                    x_start, y_start, x_end, y_end = [int(val) for val in line]
                    cv2.line(
                        img_output,
                        (x_start, y_start),
                        (x_end, y_end),
                        [255, 255, 255],
                        1,
                    )
        except Exception as e:
            pass
        return img_output[:, :, 0]


__all__ = ["MLSDDetector"]



import torch.nn as nn

from .ih_model import IHModelWithBackbone
from ..modifiers import LRMult
from ..modeling.hrnet_ocr import HighResolutionNet
from ..modeling.basic_blocks import MaxPoolDownSize


class HRNetIHModel(IHModelWithBackbone):
    def __init__(
        self,
        base_config,
        downsize_hrnet_input=False,
        mask_fusion="sum",
        lr_mult=0.1,
        cat_hrnet_outputs=True,
        pyramid_channels=-1,
        ocr=64,
        width=18,
        small=True,
        mode="cat",
        **base_kwargs
    ):
        """
        Creates image harmonization model supported by the features extracted from the pre-trained HRNet backbone.
        HRNet outputs feature maps on 4 different resolutions.

        Parameters
        ----------
        base_config : dict
            Configuration dict for the base model, to which the backbone features are incorporated.
            base_config contains model class and init parameters, examples can be found in iharm.mconfigs.base_models
        downsize_backbone_input : bool
            If the input image should be half-sized for the backbone.
        mask_fusion : str
            How to fuse the binary mask with the backbone input:
            'sum': apply convolution to the mask and sum it with the output of the first convolution in the backbone
            'rgb': concatenate the mask to the input image and translate it back to 3 channels with convolution
            otherwise: do not fuse mask with the backbone input
        lr_mult : float
            Multiply learning rate to lr_mult when updating the weights of the backbone.
        cat_hrnet_outputs : bool
            If 4 HRNet outputs should be resized and concatenated to a single tensor.
        pyramid_channels : int
            When HRNet outputs are concatenated to a single one, it can be consequently downsized
            to produce a feature pyramid.
            The pyramid features are then fused with the encoder outputs in the base model on multiple layers.
            Each pyramid feature map contains equal number of channels equal to pyramid_channels.
            If pyramid_channels <= 0, the feature pyramid is not constructed.
        ocr : int
            When HRNet outputs are concatenated to a single one, the OCR module can be applied
            resulting in feature map with (2 * ocr) channels. If ocr <= 0 the OCR module is not applied.
        width : int
            Width of the HRNet blocks.
        small : bool
            If True, HRNet contains 2 blocks at each stage and 4 otherwise.
        mode : str
            How to fuse the backbone features with the encoder outputs in the base model:
            'sum': apply convolution to the backbone feature map obtaining number of channels
             same as in the encoder output and sum them
            'cat': concatenate the backbone feature map with the encoder output
            'catc': concatenate the backbone feature map with the encoder output and apply convolution obtaining
            number of channels same as in the encoder output
            otherwise: the backbone features are not incorporated into the base model
        base_kwargs : dict
            any kwargs associated with the base model
        """
        params = base_config["params"]
        params.update(base_kwargs)
        depth = params["depth"]

        backbone = HRNetBB(
            cat_outputs=cat_hrnet_outputs,
            pyramid_channels=pyramid_channels,
            pyramid_depth=min(depth - 2 if not downsize_hrnet_input else depth - 3, 4),
            width=width,
            ocr=ocr,
            small=small,
            lr_mult=lr_mult,
        )

        params.update(
            dict(
                backbone_from=3 if downsize_hrnet_input else 2,
                backbone_channels=backbone.output_channels,
                backbone_mode=mode,
            )
        )
        base_model = base_config["model"](**params)

        super(HRNetIHModel, self).__init__(
            base_model, backbone, downsize_hrnet_input, mask_fusion
        )


class HRNetBB(nn.Module):
    def __init__(
        self,
        cat_outputs=True,
        pyramid_channels=256,
        pyramid_depth=4,
        width=18,
        ocr=64,
        small=True,
        lr_mult=0.1,
    ):
        super(HRNetBB, self).__init__()
        self.cat_outputs = cat_outputs
        self.ocr_on = ocr > 0 and cat_outputs
        self.pyramid_on = pyramid_channels > 0 and cat_outputs

        self.hrnet = HighResolutionNet(width, 2, ocr_width=ocr, small=small)
        self.hrnet.apply(LRMult(lr_mult))
        if self.ocr_on:
            self.hrnet.ocr_distri_head.apply(LRMult(1.0))
            self.hrnet.ocr_gather_head.apply(LRMult(1.0))
            self.hrnet.conv3x3_ocr.apply(LRMult(1.0))

        hrnet_cat_channels = [width * 2**i for i in range(4)]
        if self.pyramid_on:
            self.output_channels = [pyramid_channels] * 4
        elif self.ocr_on:
            self.output_channels = [ocr * 2]
        elif self.cat_outputs:
            self.output_channels = [sum(hrnet_cat_channels)]
        else:
            self.output_channels = hrnet_cat_channels

        if self.pyramid_on:
            downsize_in_channels = ocr * 2 if self.ocr_on else sum(hrnet_cat_channels)
            self.downsize = MaxPoolDownSize(
                downsize_in_channels, pyramid_channels, pyramid_channels, pyramid_depth
            )

    def forward(self, image, mask, mask_features):
        if not self.cat_outputs:
            return self.hrnet.compute_hrnet_feats(
                image, mask_features, return_list=True
            )

        outputs = list(self.hrnet(image, mask, mask_features))
        if self.pyramid_on:
            outputs = self.downsize(outputs[0])
        return outputs

    def load_pretrained_weights(self, pretrained_path):
        self.hrnet.load_pretrained_weights(pretrained_path)



import math
import torch
import numbers

import torch.nn as nn
import torch.nn.functional as F


class ConvHead(nn.Module):
    def __init__(
        self,
        out_channels,
        in_channels=32,
        num_layers=1,
        kernel_size=3,
        padding=1,
        norm_layer=nn.BatchNorm2d,
    ):
        super(ConvHead, self).__init__()
        convhead = []

        for i in range(num_layers):
            convhead.extend(
                [
                    nn.Conv2d(in_channels, in_channels, kernel_size, padding=padding),
                    nn.ReLU(),
                    norm_layer(in_channels)
                    if norm_layer is not None
                    else nn.Identity(),
                ]
            )
        convhead.append(nn.Conv2d(in_channels, out_channels, 1, padding=0))

        self.convhead = nn.Sequential(*convhead)

    def forward(self, *inputs):
        return self.convhead(inputs[0])


class SepConvHead(nn.Module):
    def __init__(
        self,
        num_outputs,
        in_channels,
        mid_channels,
        num_layers=1,
        kernel_size=3,
        padding=1,
        dropout_ratio=0.0,
        dropout_indx=0,
        norm_layer=nn.BatchNorm2d,
    ):
        super(SepConvHead, self).__init__()

        sepconvhead = []

        for i in range(num_layers):
            sepconvhead.append(
                SeparableConv2d(
                    in_channels=in_channels if i == 0 else mid_channels,
                    out_channels=mid_channels,
                    dw_kernel=kernel_size,
                    dw_padding=padding,
                    norm_layer=norm_layer,
                    activation="relu",
                )
            )
            if dropout_ratio > 0 and dropout_indx == i:
                sepconvhead.append(nn.Dropout(dropout_ratio))

        sepconvhead.append(
            nn.Conv2d(
                in_channels=mid_channels,
                out_channels=num_outputs,
                kernel_size=1,
                padding=0,
            )
        )

        self.layers = nn.Sequential(*sepconvhead)

    def forward(self, *inputs):
        x = inputs[0]

        return self.layers(x)


def select_activation_function(activation):
    if isinstance(activation, str):
        if activation.lower() == "relu":
            return nn.ReLU
        elif activation.lower() == "softplus":
            return nn.Softplus
        else:
            raise ValueError(f"Unknown activation type {activation}")
    elif isinstance(activation, nn.Module):
        return activation
    else:
        raise ValueError(f"Unknown activation type {activation}")


class SeparableConv2d(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        dw_kernel,
        dw_padding,
        dw_stride=1,
        activation=None,
        use_bias=False,
        norm_layer=None,
    ):
        super(SeparableConv2d, self).__init__()
        _activation = select_activation_function(activation)
        self.body = nn.Sequential(
            nn.Conv2d(
                in_channels,
                in_channels,
                kernel_size=dw_kernel,
                stride=dw_stride,
                padding=dw_padding,
                bias=use_bias,
                groups=in_channels,
            ),
            nn.Conv2d(
                in_channels, out_channels, kernel_size=1, stride=1, bias=use_bias
            ),
            norm_layer(out_channels) if norm_layer is not None else nn.Identity(),
            _activation(),
        )

    def forward(self, x):
        return self.body(x)


class ConvBlock(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size=4,
        stride=2,
        padding=1,
        norm_layer=nn.BatchNorm2d,
        activation=nn.ELU,
        bias=True,
    ):
        super(ConvBlock, self).__init__()
        self.block = nn.Sequential(
            nn.Conv2d(
                in_channels,
                out_channels,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                bias=bias,
            ),
            norm_layer(out_channels) if norm_layer is not None else nn.Identity(),
            activation(),
        )

    def forward(self, x):
        return self.block(x)


class GaussianSmoothing(nn.Module):
    """
    https://discuss.pytorch.org/t/is-there-anyway-to-do-gaussian-filtering-for-an-image-2d-3d-in-pytorch/12351/10
    Apply gaussian smoothing on a tensor (1d, 2d, 3d).
    Filtering is performed seperately for each channel in the input using a depthwise convolution.
    Arguments:
        channels (int, sequence): Number of channels of the input tensors.
            Output will have this number of channels as well.
        kernel_size (int, sequence): Size of the gaussian kernel.
        sigma (float, sequence): Standard deviation of the gaussian kernel.
        dim (int, optional): The number of dimensions of the data. Default value is 2 (spatial).
    """

    def __init__(self, channels, kernel_size, sigma, padding=0, dim=2):
        super(GaussianSmoothing, self).__init__()
        if isinstance(kernel_size, numbers.Number):
            kernel_size = [kernel_size] * dim
        if isinstance(sigma, numbers.Number):
            sigma = [sigma] * dim

        # The gaussian kernel is the product of the gaussian function of each dimension.
        kernel = 1.0
        meshgrids = torch.meshgrid(
            [torch.arange(size, dtype=torch.float32) for size in kernel_size]
        )
        for size, std, grid in zip(kernel_size, sigma, meshgrids):
            mean = (size - 1) / 2.0
            kernel *= torch.exp(-(((grid - mean) / std) ** 2) / 2) / (
                std * (2 * math.pi) ** 0.5
            )
        # Make sure sum of values in gaussian kernel equals 1.
        kernel = kernel / torch.sum(kernel)
        # Reshape to depthwise convolutional weight.
        kernel = kernel.view(1, 1, *kernel.size())
        kernel = torch.repeat_interleave(kernel, channels, 0)

        self.register_buffer("weight", kernel)
        self.groups = channels
        self.padding = padding

        if dim == 1:
            self.conv = F.conv1d
        elif dim == 2:
            self.conv = F.conv2d
        elif dim == 3:
            self.conv = F.conv3d
        else:
            raise RuntimeError(
                "Only 1, 2 and 3 dimensions are supported. Received {}.".format(dim)
            )

    def forward(self, input):
        """
        Apply gaussian filter to input.
        Arguments:
            input (torch.Tensor): Input to apply gaussian filter on.
        Returns:
            filtered (torch.Tensor): Filtered output.
        """
        return self.conv(
            input, weight=self.weight, padding=self.padding, groups=self.groups
        )


class MaxPoolDownSize(nn.Module):
    def __init__(self, in_channels, mid_channels, out_channels, depth):
        super(MaxPoolDownSize, self).__init__()
        self.depth = depth
        self.reduce_conv = ConvBlock(
            in_channels, mid_channels, kernel_size=1, stride=1, padding=0
        )
        self.convs = nn.ModuleList(
            [
                ConvBlock(
                    mid_channels, out_channels, kernel_size=3, stride=1, padding=1
                )
                for conv_i in range(depth)
            ]
        )
        self.pool2d = nn.MaxPool2d(kernel_size=2)

    def forward(self, x):
        outputs = []

        output = self.reduce_conv(x)

        for conv_i, conv in enumerate(self.convs):
            output = output if conv_i == 0 else self.pool2d(output)
            outputs.append(conv(output))

        return outputs



import torch.nn as nn

from torch import Tensor
from typing import Dict
from typing import List
from typing import Optional
from cftool.misc import safe_execute

from .base import MLModel
from ...schema import MLEncoderSettings
from ...schema import MLGlobalEncoderSettings
from ...modules.blocks import mapping_dict


@MLModel.register("fcnn")
class FCNN(MLModel):
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        num_history: int = 1,
        hidden_units: Optional[List[int]] = None,
        *,
        mapping_type: str = "basic",
        bias: bool = True,
        activation: str = "ReLU",
        batch_norm: bool = False,
        dropout: float = 0.0,
        rank: Optional[int] = None,
        rank_ratio: Optional[float] = None,
        encoder_settings: Optional[Dict[str, MLEncoderSettings]] = None,
        global_encoder_settings: Optional[MLGlobalEncoderSettings] = None,
    ):
        super().__init__(
            encoder_settings=encoder_settings,
            global_encoder_settings=global_encoder_settings,
        )
        if self.encoder is not None:
            input_dim += self.encoder.dim_increment
        input_dim *= num_history
        if hidden_units is None:
            dim = max(32, min(1024, 2 * input_dim))
            hidden_units = 2 * [dim]
        mapping_base = mapping_dict[mapping_type]
        blocks: List[nn.Module] = []
        for hidden_unit in hidden_units:
            mapping = safe_execute(
                mapping_base,
                dict(
                    in_dim=input_dim,
                    out_dim=hidden_unit,
                    bias=bias,
                    activation=activation,
                    batch_norm=batch_norm,
                    dropout=dropout,
                    rank=rank,
                    rank_ratio=rank_ratio,
                ),
            )
            blocks.append(mapping)
            input_dim = hidden_unit
        blocks.append(nn.Linear(input_dim, output_dim, bias))
        self.hidden_units = hidden_units
        self.net = nn.Sequential(*blocks)

    def forward(self, net: Tensor) -> Tensor:
        if len(net.shape) > 2:
            net = net.contiguous().view(len(net), -1)
        net = self.encode(net).merged_all
        return self.net(net)


__all__ = ["FCNN"]



from abc import abstractmethod
from abc import ABC

from ..schema import IDataLoader
from ..schema import MetricsOutputs


class IEvaluationPipeline(ABC):
    @abstractmethod
    def evaluate(self, loader: IDataLoader) -> MetricsOutputs:
        pass


__all__ = [
    "IEvaluationPipeline",
]



import numpy as np

from abc import abstractmethod
from abc import ABC
from typing import Any

from .schema import IEvaluationPipeline
from .blocks import BuildMetricsBlock
from ..schema import DLConfig
from ..schema import IDataLoader
from ..schema import MetricsOutputs
from ..constants import INPUT_KEY
from ..constants import PREDICTIONS_KEY


class IPredictor(ABC):
    @abstractmethod
    def predict(self, x: np.ndarray) -> np.ndarray:
        pass


class SKLearnClassifier(IPredictor):
    def __init__(self, m: Any) -> None:
        self.m = m

    def predict(self, x: np.ndarray) -> np.ndarray:
        return self.m.predict_log_proba(x)


class GeneralEvaluationPipeline(IEvaluationPipeline):
    def __init__(self, config: DLConfig, predictor: IPredictor) -> None:
        b_metrics = BuildMetricsBlock()
        b_metrics.build(config)
        if b_metrics.metrics is None:
            raise ValueError(
                "`metrics` should not be `None` for `GeneralPredictor`, "
                "you may try specifying `metric_names` in `config`"
            )
        self.m = predictor
        self.metrics = b_metrics.metrics

    def evaluate(self, loader: IDataLoader) -> MetricsOutputs:
        full_batch = loader.get_full_batch()
        predictions = self.m.predict(full_batch[INPUT_KEY])
        return self.metrics.evaluate(full_batch, {PREDICTIONS_KEY: predictions}, loader)


__all__ = [
    "IPredictor",
    "SKLearnClassifier",
    "GeneralEvaluationPipeline",
]



import torch

import numpy as np
import torch.nn as nn

from torch import Tensor
from typing import Any
from typing import List
from typing import Tuple
from typing import Optional
from torchvision.models import vgg16
from torchvision.models import vgg19

from ..register import register_backbone
from ......misc.toolkit import download_model
from ......modules.blocks import SEBlock


register_backbone(
    "vgg16",
    [64, 128, 256, 512],
    {
        "features.3": "stage0",
        "features.8": "stage1",
        "features.15": "stage2",
        "features.22": "stage3",
    },
)(vgg16)


register_backbone(
    "vgg16_full",
    [64, 128, 256, 512, 512],
    {
        "features.3": "stage0",
        "features.8": "stage1",
        "features.15": "stage2",
        "features.22": "stage3",
        "features.29": "stage4",
    },
)(vgg16)


register_backbone(
    "vgg19",
    [64, 128, 256, 512],
    {
        "features.3": "stage0",
        "features.8": "stage1",
        "features.17": "stage2",
        "features.26": "stage3",
    },
)(vgg19)


register_backbone(
    "vgg19_lite",
    [64, 128, 256, 512],
    {
        "features.1": "stage0",
        "features.6": "stage1",
        "features.11": "stage2",
        "features.20": "stage3",
    },
)(vgg19)


register_backbone(
    "vgg19_large",
    [64, 128, 256, 512, 512, 512],
    {
        "features.3": "stage0",
        "features.8": "stage1",
        "features.17": "stage2",
        "features.22": "stage3_first",
        "features.26": "stage3_second",
        "features.35": "stage4",
    },
)(vgg19)


@register_backbone(
    "vgg_style",
    [64, 128, 256, 512],
    {
        "net.3": "stage0",
        "net.10": "stage1",
        "net.17": "stage2",
        "net.30": "stage3",
    },
)
class VGGStyle(nn.Module):
    def __init__(self, pretrained: bool = True):
        super().__init__()
        self.net = nn.Sequential(
            nn.Conv2d(3, 3, (1, 1)),
            nn.ReflectionPad2d((1, 1, 1, 1)),
            nn.Conv2d(3, 64, (3, 3)),
            nn.ReLU(),
            nn.ReflectionPad2d((1, 1, 1, 1)),
            nn.Conv2d(64, 64, (3, 3)),
            nn.ReLU(),
            nn.MaxPool2d((2, 2), (2, 2), (0, 0), ceil_mode=True),
            nn.ReflectionPad2d((1, 1, 1, 1)),
            nn.Conv2d(64, 128, (3, 3)),
            nn.ReLU(),
            nn.ReflectionPad2d((1, 1, 1, 1)),
            nn.Conv2d(128, 128, (3, 3)),
            nn.ReLU(),
            nn.MaxPool2d((2, 2), (2, 2), (0, 0), ceil_mode=True),
            nn.ReflectionPad2d((1, 1, 1, 1)),
            nn.Conv2d(128, 256, (3, 3)),
            nn.ReLU(),
            nn.ReflectionPad2d((1, 1, 1, 1)),
            nn.Conv2d(256, 256, (3, 3)),
            nn.ReLU(),
            nn.ReflectionPad2d((1, 1, 1, 1)),
            nn.Conv2d(256, 256, (3, 3)),
            nn.ReLU(),
            nn.ReflectionPad2d((1, 1, 1, 1)),
            nn.Conv2d(256, 256, (3, 3)),
            nn.ReLU(),
            nn.MaxPool2d((2, 2), (2, 2), (0, 0), ceil_mode=True),
            nn.ReflectionPad2d((1, 1, 1, 1)),
            nn.Conv2d(256, 512, (3, 3)),
            nn.ReLU(),
        )
        if pretrained:
            pt_path = download_model("vgg_style")
            self.load_state_dict(torch.load(pt_path))

    def forward(self, net: torch.Tensor) -> torch.Tensor:
        return self.net(net)


class RepVGGBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        *,
        deploy: bool,
        stride: int = 1,
        dilation: int = 1,
        groups: int = 1,
        use_post_se: bool = True,
    ):
        super().__init__()

        kernel_size = self.kernel_size = 3
        padding = self.padding = 1

        self.deploy = deploy
        self.stride = stride
        self.dilation = dilation
        self.groups = groups
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.relu = nn.ReLU(inplace=True)

        if not use_post_se:
            self.post_se = nn.Identity()
        else:
            self.post_se = SEBlock(out_channels, out_channels // 4, block_impl="torch")

        if deploy:
            self.conv_fused = nn.Conv2d(
                in_channels,
                out_channels,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                dilation=dilation,
                groups=groups,
                bias=True,
            )
        else:
            if out_channels == in_channels and stride == 1:
                self.identity = nn.BatchNorm2d(num_features=out_channels)
            else:
                self.identity = None
            self.dense = nn.Sequential(
                nn.Conv2d(
                    in_channels,
                    out_channels,
                    kernel_size,
                    stride,
                    padding,
                    groups=groups,
                    bias=False,
                ),
                nn.BatchNorm2d(out_channels),
            )
            self.side = nn.Sequential(
                nn.Conv2d(
                    in_channels,
                    out_channels,
                    1,
                    stride,
                    padding - kernel_size // 2,
                    groups=groups,
                    bias=False,
                ),
                nn.BatchNorm2d(out_channels),
            )

    def forward(self, net: Tensor) -> Tensor:
        if self.deploy:
            return self.post_se(self.relu(self.conv_fused(net)))
        if self.identity is None:
            id_out = 0
        else:
            id_out = self.identity(net)
        out = self.dense(net) + self.side(net) + id_out
        return self.post_se(self.relu(out))

    def get_equivalent_kernel_bias(self) -> Tuple[Tensor, Tensor]:
        kernel3x3, bias3x3 = self._fuse_bn_tensor(self.dense)
        kernel1x1, bias1x1 = self._fuse_bn_tensor(self.side)
        kernel_id, bias_id = self._fuse_bn_tensor(self.identity)
        kernel1x1_padded = nn.functional.pad(kernel1x1, [1, 1, 1, 1])
        kernel = kernel3x3 + kernel1x1_padded + kernel_id
        bias = bias3x3 + bias1x1 + bias_id
        return kernel, bias

    def _fuse_bn_tensor(self, branch: Optional[nn.Module]) -> Any:
        if branch is None:
            return 0, 0
        if isinstance(branch, nn.Sequential):
            conv, bn = branch[0], branch[1]
            kernel, running_mean, running_var, gamma, beta, eps = (
                conv.weight,
                bn.running_mean,
                bn.running_var,
                bn.weight,
                bn.bias,
                bn.eps,
            )
        else:
            assert isinstance(branch, nn.BatchNorm2d)
            if not hasattr(self, "id_tensor"):
                input_dim = self.in_channels // self.groups
                kernel_value = np.zeros(
                    [self.in_channels, input_dim, 3, 3],
                    dtype=np.float32,
                )
                for i in range(self.in_channels):
                    kernel_value[i, i % input_dim, 1, 1] = 1
                self.id_tensor = torch.from_numpy(kernel_value).to(branch.weight.device)
            kernel, running_mean, running_var, gamma, beta, eps = (
                self.id_tensor,
                branch.running_mean,
                branch.running_var,
                branch.weight,
                branch.bias,
                branch.eps,
            )
        std = (running_var + eps).sqrt()
        t = (gamma / std).reshape(-1, 1, 1, 1)
        return kernel * t, beta - running_mean * gamma / std

    def switch_to_deploy(self) -> None:
        if hasattr(self, "conv_fused"):
            return None
        kernel, bias = self.get_equivalent_kernel_bias()
        self.conv_fused = nn.Conv2d(
            in_channels=self.in_channels,
            out_channels=self.out_channels,
            kernel_size=self.kernel_size,
            stride=self.stride,
            padding=self.padding,
            dilation=self.dilation,
            groups=self.groups,
            bias=True,
        )
        self.conv_fused.weight.data = kernel
        self.conv_fused.bias.data = bias  # type: ignore
        self.__delattr__("dense")
        self.__delattr__("side")
        if hasattr(self, "identity"):
            self.__delattr__("identity")
        if hasattr(self, "id_tensor"):
            self.__delattr__("id_tensor")
        self.deploy = True


class RepVGGStage(nn.Module):
    def __init__(
        self,
        in_channels: int,
        latent_channels: int,
        num_blocks: int,
        *,
        deploy: bool = False,
        stride: int = 1,
        dilation: int = 1,
        use_post_se: bool = True,
    ):
        super().__init__()
        strides = [stride] + [1] * (num_blocks - 1)
        blocks = []
        in_nc = in_channels
        out_nc = latent_channels
        for stride in strides:
            blocks.append(
                RepVGGBlock(
                    in_nc,
                    out_nc,
                    deploy=deploy,
                    stride=stride,
                    dilation=dilation,
                    groups=1,
                    use_post_se=use_post_se,
                )
            )
            in_nc = out_nc
        self.net = nn.Sequential(*blocks)

    def forward(self, net: Tensor) -> Tensor:
        return self.net(net)

    def switch_to_deploy(self) -> None:
        for block in self.net:
            block.switch_to_deploy()


class RepVGG(nn.Module):
    def __init__(
        self,
        num_blocks: List[int],
        width_multiplier: List[float],
        *,
        use_post_se: bool = True,
    ):
        super().__init__()
        deploy = False
        out_channels = min(64, int(64 * width_multiplier[0]))
        self.stage1 = RepVGGBlock(
            3,
            out_channels,
            deploy=deploy,
            stride=2,
            use_post_se=use_post_se,
        )
        self.stage2 = RepVGGStage(
            out_channels,
            int(64 * width_multiplier[0]),
            num_blocks[0],
            deploy=deploy,
            stride=2,
            use_post_se=use_post_se,
        )
        self.stage3 = RepVGGStage(
            int(64 * width_multiplier[0]),
            int(128 * width_multiplier[1]),
            num_blocks[1],
            deploy=deploy,
            stride=2,
            use_post_se=use_post_se,
        )
        self.stage4_first = RepVGGStage(
            int(128 * width_multiplier[1]),
            int(256 * width_multiplier[2]),
            num_blocks[2] // 2,
            deploy=deploy,
            stride=2,
            use_post_se=use_post_se,
        )
        self.stage4_second = RepVGGStage(
            int(256 * width_multiplier[2]),
            int(256 * width_multiplier[2]),
            num_blocks[2] // 2,
            deploy=deploy,
            stride=1,
            use_post_se=use_post_se,
        )
        self.stage5 = RepVGGStage(
            int(256 * width_multiplier[2]),
            int(512 * width_multiplier[3]),
            num_blocks[3],
            deploy=deploy,
            stride=2,
            use_post_se=use_post_se,
        )

    def forward(self, net: Tensor) -> Tensor:
        net = self.stage1(net)
        net = self.stage2(net)
        net = self.stage3(net)
        net = self.stage4_first(net)
        net = self.stage4_second(net)
        net = self.stage5(net)
        return net

    def switch_to_deploy(self) -> None:
        self.stage1.switch_to_deploy()
        self.stage2.switch_to_deploy()
        self.stage3.switch_to_deploy()
        self.stage4_first.switch_to_deploy()
        self.stage4_second.switch_to_deploy()
        self.stage5.switch_to_deploy()


@register_backbone(
    "rep_vgg",
    [64, 128, 256, 512, 512, 2048],
    dict(
        stage1="stage1",
        stage2="stage2",
        stage3="stage3",
        stage4_first="stage4_first",
        stage4_second="stage4_second",
        stage5="stage5",
    ),
)
def rep_vgg(pretrained: bool = False) -> RepVGG:
    if pretrained:
        raise ValueError("`RepVGG` does not support `pretrained`")
    return RepVGG([4, 6, 16, 1], [2.0, 2.0, 2.0, 4.0])


@register_backbone(
    "rep_vgg_lite",
    [48, 48, 96, 192, 192, 1280],
    dict(
        stage1="stage1",
        stage2="stage2",
        stage3="stage3",
        stage4_first="stage4_first",
        stage4_second="stage4_second",
        stage5="stage5",
    ),
)
def rep_vgg_lite(pretrained: bool = False) -> RepVGG:
    if pretrained:
        raise ValueError("`RepVGG` does not support `pretrained`")
    return RepVGG([2, 4, 14, 1], [0.75, 0.75, 0.75, 2.5])


@register_backbone(
    "rep_vgg_large",
    [64, 160, 320, 640, 640, 2560],
    dict(
        stage1="stage1",
        stage2="stage2",
        stage3="stage3",
        stage4_first="stage4_first",
        stage4_second="stage4_second",
        stage5="stage5",
    ),
)
def rep_vgg_large(pretrained: bool = False) -> RepVGG:
    if pretrained:
        raise ValueError("`RepVGG` does not support `pretrained`")
    return RepVGG([8, 14, 24, 1], [2.5, 2.5, 2.5, 5.0])


__all__ = [
    "rep_vgg",
    "rep_vgg_lite",
    "rep_vgg_large",
]



import torch
from functools import partial

from torch import nn as nn

from ..ops import ChannelAttention
from ..modeling.unet import UNetDecoder
from ..modeling.unet import UNetEncoder
from ..modeling.basic_blocks import ConvBlock
from ..modeling.basic_blocks import GaussianSmoothing


class SSAMImageHarmonization(nn.Module):
    def __init__(
        self,
        depth,
        norm_layer=nn.BatchNorm2d,
        batchnorm_from=2,
        attend_from=3,
        attention_mid_k=2.0,
        image_fusion=False,
        ch=64,
        max_channels=512,
        backbone_from=-1,
        backbone_channels=None,
        backbone_mode="",
    ):
        super(SSAMImageHarmonization, self).__init__()
        self.depth = depth
        self.encoder = UNetEncoder(
            depth,
            ch,
            norm_layer,
            batchnorm_from,
            max_channels,
            backbone_from,
            backbone_channels,
            backbone_mode,
        )
        self.decoder = UNetDecoder(
            depth,
            self.encoder.block_channels,
            norm_layer,
            attention_layer=partial(SpatialSeparatedAttention, mid_k=attention_mid_k),
            attend_from=attend_from,
            image_fusion=image_fusion,
        )

    def forward(self, image, mask, backbone_features=None):
        x = torch.cat((image, mask), dim=1)
        intermediates = self.encoder(x, backbone_features)
        output = self.decoder(intermediates, image, mask)
        return {"images": output}


class SpatialSeparatedAttention(nn.Module):
    def __init__(self, in_channels, norm_layer, activation, mid_k=2.0):
        super(SpatialSeparatedAttention, self).__init__()
        self.background_gate = ChannelAttention(in_channels)
        self.foreground_gate = ChannelAttention(in_channels)
        self.mix_gate = ChannelAttention(in_channels)

        mid_channels = int(mid_k * in_channels)
        self.learning_block = nn.Sequential(
            ConvBlock(
                in_channels,
                mid_channels,
                kernel_size=3,
                stride=1,
                padding=1,
                norm_layer=norm_layer,
                activation=activation,
                bias=False,
            ),
            ConvBlock(
                mid_channels,
                in_channels,
                kernel_size=3,
                stride=1,
                padding=1,
                norm_layer=norm_layer,
                activation=activation,
                bias=False,
            ),
        )
        self.mask_blurring = GaussianSmoothing(1, 7, 1, padding=3)

    def forward(self, x, mask):
        mask = self.mask_blurring(
            nn.functional.interpolate(
                mask, size=x.size()[-2:], mode="bilinear", align_corners=True
            )
        )
        background = self.background_gate(x)
        foreground = self.learning_block(self.foreground_gate(x))
        mix = self.mix_gate(x)
        output = mask * (foreground + mix) + (1 - mask) * background
        return output



import re
import dill
import math
import uuid
import torch

from torch import Tensor
from typing import Any
from typing import Dict
from typing import List
from typing import NamedTuple

from .schema import IConditionModel
from ....nlp.tokenizers import ITokenizer
from ....nlp.tokenizers import ICLIPTokenizer
from ....multimodal.clip import CLIP
from .....misc.toolkit import get_device


weight_pattern = re.compile(
    r"""
\\\(|
\\\)|
\\\[|
\\]|
\\\\|
\\|
\(|
\[|
:([+-]?[.\d]+)\)|
\)|
]|
[^\\()\[\]:]+|
:
""",
    re.X,
)


def parse_weights(text: str) -> List[List[Any]]:
    res: List[List[Any]] = []
    round_brackets = []
    square_brackets = []

    round_bracket_multiplier = 1.1
    square_bracket_multiplier = 1 / 1.1

    def multiply_range(start_position: int, multiplier: float) -> None:
        for p in range(start_position, len(res)):
            res[p][1] *= multiplier

    for m in weight_pattern.finditer(text):
        text = m.group(0)
        weight = m.group(1)

        if text.startswith("\\"):
            res.append([text[1:], 1.0])
        elif text == "(":
            round_brackets.append(len(res))
        elif text == "[":
            square_brackets.append(len(res))
        elif weight is not None and len(round_brackets) > 0:
            multiply_range(round_brackets.pop(), float(weight))
        elif text == ")" and len(round_brackets) > 0:
            multiply_range(round_brackets.pop(), round_bracket_multiplier)
        elif text == "]" and len(square_brackets) > 0:
            multiply_range(square_brackets.pop(), square_bracket_multiplier)
        else:
            res.append([text, 1.0])

    for pos in round_brackets:
        multiply_range(pos, round_bracket_multiplier)

    for pos in square_brackets:
        multiply_range(pos, square_bracket_multiplier)

    if len(res) == 0:
        res = [["", 1.0]]

    i = 0
    while i + 1 < len(res):
        if res[i][1] != res[i + 1][1]:
            i += 1
        else:
            res[i][0] += res[i + 1][0]
            res.pop(i + 1)

    return res


class CustomToken(NamedTuple):
    name: str
    uuid: str
    token_id: int
    embedding: Tensor


class inject_embeddings:
    def __init__(self, cond_model: "CLIPTextConditionModel") -> None:
        self.m = cond_model
        self.original_embeddings = cond_model.embeddings
        self.changed = False

    def __enter__(self) -> None:
        sorted_keys = sorted(self.m.customized)
        if not sorted_keys:
            return
        anchor = self.original_embeddings
        self.changed = True
        new_inject = [self.m.customized[k].embedding.to(anchor) for k in sorted_keys]
        start = len(anchor)
        for i, k in enumerate(sorted_keys):
            if start + i != k:
                msg = "injected embeddings not contiguous to the original embeddings"
                raise ValueError(msg)
        concat = torch.cat([anchor] + new_inject, dim=0)
        self._inject(concat)

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        if not self.changed:
            return
        self._inject(self.original_embeddings)

    def _inject(self, embeddings: Tensor) -> None:
        if self.m.m.token_embedding is None:
            raise ValueError("`token_embedding` is None")
        self.m.m.token_embedding.weight.data = embeddings


@IConditionModel.register("multimodal/clip")
@IConditionModel.register("multimodal/clip.large")
@IConditionModel.register("multimodal/clip.chinese")
@IConditionModel.register("multimodal/clip.open_clip_ViT_H_14")
class CLIPTextConditionModel(IConditionModel):
    m: CLIP
    tokenizer: ICLIPTokenizer

    def __init__(self, m: CLIP):
        super().__init__(m)
        self.clip_skip = 0
        self.context_length = m.context_length
        tokenizer = "clip.chinese" if m.context_length == 512 else "clip"
        self.tokenizer = ITokenizer.make(tokenizer, {})
        self.comma_token = self.tokenizer.tokenizer.get_vocab()[",</w>"]
        self.comma_padding_backtrack = 20
        self._dumped_tokenizer = dill.dumps(self.tokenizer.tokenizer)
        self.dictionary: Dict[str, str] = {}
        self.customized: Dict[int, CustomToken] = {}

    @property
    def embeddings(self) -> Tensor:
        if self.m.token_embedding is None:
            raise ValueError("`token_embedding` is None")
        return self.m.token_embedding.weight.data

    def register_custom(self, embeddings: Dict[str, List[List[float]]]) -> None:
        existing = self.embeddings
        dtype = existing.dtype
        device = existing.device
        for name, embedding in embeddings.items():
            tensor = torch.asarray(embedding, dtype=dtype, device=device)
            embedding_dim = self.embeddings.shape[1]
            if tensor.shape[1] != embedding_dim:
                raise ValueError(
                    f"dimension of the custom embedding '{name}' is not correct "
                    f"(expected {embedding_dim}, got {tensor.shape[1]})"
                )
            tensor = tensor.view(-1, tensor.shape[-1])
            tags = []
            for i in range(tensor.shape[0]):
                i_tensor = tensor[[i]]
                tag = uuid.uuid4().hex
                self.tokenizer.tokenizer.add_tokens(tag)
                token_id = self.tokenizer.tokenizer.convert_tokens_to_ids(tag)
                self.customized[token_id] = CustomToken(name, tag, token_id, i_tensor)
                tags.append(tag)
            self.dictionary[name] = " ".join(tags)

    def clear_custom(self) -> None:
        self.dictionary = {}
        self.customized = {}
        # TODO : optimize this
        self.tokenizer.tokenizer = dill.loads(self._dumped_tokenizer)

    def get_line(self, text: str, clip_skip: int) -> Tensor:
        for k, v in self.dictionary.items():
            text = text.replace(k, v)
        parsed = parse_weights(text)
        parsed_texts = [pair[0] for pair in parsed]
        parsed_weights = [pair[1] for pair in parsed]
        tokenized = self.tokenizer.encode(
            parsed_texts,
            truncation=False,
            add_special_tokens=False,
        )["input_ids"]

        concat_ids: List[int] = []
        weights: List[float] = []
        id_end = self.tokenizer.eos_token_id
        last_comma = -1
        valid_context_length = self.context_length - 2
        for tokens, weight in zip(tokenized, parsed_weights):
            i = 0
            while i < len(tokens):
                token = tokens[i]
                if token == self.comma_token:
                    last_comma = len(concat_ids)
                elif (
                    self.comma_padding_backtrack != 0
                    and max(len(concat_ids), 1) % valid_context_length == 0
                    and last_comma != -1
                    and len(concat_ids) - last_comma <= self.comma_padding_backtrack
                ):
                    last_comma += 1
                    reloc_tokens = concat_ids[last_comma:]
                    reloc_mults = weights[last_comma:]

                    concat_ids = concat_ids[:last_comma]
                    length = len(concat_ids)

                    rem = (
                        int(math.ceil(length / valid_context_length))
                        * valid_context_length
                        - length
                    )
                    concat_ids += [id_end] * rem + reloc_tokens
                    weights = weights[:last_comma] + [1.0] * rem + reloc_mults

                concat_ids.append(token)
                weights.append(weight)
                i += 1

        zs = []
        is_first = True
        remained_ids = concat_ids
        remained_weights = weights
        while is_first or remained_ids:
            local_ids = remained_ids[:]
            local_weights = remained_weights[:]
            local_ids = [self.tokenizer.bos_token_id] + local_ids
            local_weights = [1.0] + local_weights
            # padding
            diff = self.context_length - len(local_ids)
            if diff > 0:
                local_ids += [self.tokenizer.eos_token_id] * diff
                local_weights += [1.0] * diff
                remained_ids = []
            else:
                remained_ids = local_ids[self.context_length - 1 :]
                remained_weights = local_weights[self.context_length - 1 :]
                local_ids = local_ids[: self.context_length - 1]
                local_weights = local_weights[: self.context_length - 1]
                local_ids.append(self.tokenizer.eos_token_id)
                local_weights.append(1.0)
            # encode
            to_torch = lambda l, dtype: torch.asarray(
                l,
                dtype=dtype,
                device=get_device(self.m),
            )
            inp = to_torch([local_ids], torch.int64)
            weights_tensor = to_torch(local_weights, torch.float32).view(1, -1, 1)
            with inject_embeddings(self):
                z = self.m.encode_text(
                    inp,
                    apply_pooling=False,
                    determinate=False,
                    clip_skip=clip_skip,
                )
            # weighting
            original_mean = z.mean()
            z *= weights_tensor
            new_mean = z.mean()
            z *= original_mean / new_mean
            zs.append(z)
            # set flag
            is_first = False

        if len(zs) == 1:
            return zs[0]
        return torch.cat(zs, dim=1)

    def forward(self, cond: List[str]) -> Tensor:
        lines = [self.get_line(text, self.clip_skip) for text in cond]
        return torch.cat(lines, dim=0)


__all__ = [
    "CLIPTextConditionModel",
]



import torch

import torch.nn as nn

from abc import abstractmethod
from tqdm import tqdm
from torch import Tensor
from typing import Any
from typing import Dict
from typing import Type
from typing import Callable
from typing import Optional
from typing import Protocol
from cftool.misc import update_dict
from cftool.misc import safe_execute
from cftool.misc import shallow_copy_dict
from cftool.misc import WithRegister

from ..utils import cond_type
from ..utils import extract_to
from ..utils import get_timesteps
from ..utils import CONCAT_KEY
from ..utils import CONTROL_HINT_KEY
from ..utils import CONTROL_HINT_START_KEY


samplers: Dict[str, Type["ISampler"]] = {}


def is_misc_key(key: str) -> bool:
    return key in (CONCAT_KEY, CONTROL_HINT_KEY, CONTROL_HINT_START_KEY)


class Denoise(Protocol):
    def __call__(
        self,
        image: Tensor,
        timesteps: Tensor,
        cond: Optional[cond_type],
        step: int,
        total_step: int,
    ) -> Tensor:
        pass


class IDiffusion:
    _get_cond: Callable
    predict_eps_from_z_and_v: Callable
    predict_start_from_z_and_v: Callable

    denoise: Denoise
    q_sampler: "DDPMQSampler"
    condition_model: Optional[nn.Module]
    first_stage: nn.Module

    t: int
    parameterization: str
    posterior_coef1: Tensor
    posterior_coef2: Tensor
    posterior_log_variance_clipped: Tensor

    betas: Tensor
    alphas_cumprod: Tensor
    alphas_cumprod_prev: Tensor


class IQSampler:
    def __init__(self, model: IDiffusion):
        self.model = model

    @abstractmethod
    def q_sample(
        self,
        net: Tensor,
        timesteps: Tensor,
        noise: Optional[Tensor] = None,
    ) -> Tensor:
        pass

    @abstractmethod
    def reset_buffers(self, **kwargs: Any) -> None:
        pass


class DDPMQSampler(IQSampler):
    sqrt_alphas: Tensor
    sqrt_one_minus_alphas: Tensor

    def q_sample(
        self,
        net: Tensor,
        timesteps: Tensor,
        noise: Optional[Tensor] = None,
    ) -> Tensor:
        self.sqrt_alphas = self.sqrt_alphas.to(net)
        self.sqrt_one_minus_alphas = self.sqrt_one_minus_alphas.to(net)
        num_dim = len(net.shape)
        w_net = extract_to(self.sqrt_alphas, timesteps, num_dim)
        w_noise = extract_to(self.sqrt_one_minus_alphas, timesteps, num_dim)
        if noise is None:
            noise = torch.randn_like(net)
        net = w_net * net + w_noise * noise
        return net

    def reset_buffers(self, sqrt_alpha: Tensor, sqrt_one_minus_alpha: Tensor) -> None:  # type: ignore
        self.sqrt_alphas = sqrt_alpha
        self.sqrt_one_minus_alphas = sqrt_one_minus_alpha


class ISampler(WithRegister):
    d = samplers

    default_steps: int

    def __init__(self, model: IDiffusion):
        self.model = model
        self.initialized = False

    @property
    @abstractmethod
    def q_sampler(self) -> IQSampler:
        pass

    @property
    @abstractmethod
    def sample_kwargs(self) -> Dict[str, Any]:
        pass

    @abstractmethod
    def sample_step(
        self,
        image: Tensor,
        cond: Optional[cond_type],
        step: int,
        total_step: int,
        **kwargs: Any,
    ) -> Tensor:
        pass

    def q_sample(
        self,
        net: Tensor,
        timesteps: Tensor,
        noise: Optional[Tensor] = None,
    ) -> Tensor:
        return self.q_sampler.q_sample(net, timesteps, noise)

    def sample(
        self,
        z: Tensor,
        *,
        ref: Optional[Tensor] = None,
        ref_mask: Optional[Tensor] = None,
        ref_noise: Optional[Tensor] = None,
        cond: Optional[Any] = None,
        num_steps: Optional[int] = None,
        start_step: Optional[int] = None,
        verbose: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        # setup
        if num_steps is None:
            num_steps = getattr(self, "default_steps", self.model.t)
            assert isinstance(num_steps, int)
        if start_step is None:
            start_step = 0
        iterator = list(range(start_step, num_steps))
        if verbose:
            iterator = tqdm(iterator, desc=f"sampling ({self.__identifier__})")
        # execute
        image = z
        if cond is not None and self.model.condition_model is not None:
            cond = self.model._get_cond(cond)
        for step in iterator:
            callback = kwargs.get("step_callback")
            if callback is not None:
                callback_kw = dict(step=step, num_steps=num_steps, image=image)
                if not safe_execute(callback, callback_kw):
                    break
            kw = shallow_copy_dict(self.sample_kwargs)
            update_dict(shallow_copy_dict(kwargs), kw)
            image = self.sample_step(image, cond, step, num_steps, **kw)
            if ref is not None and ref_mask is not None and ref_noise is not None:
                ref_ts = get_timesteps(num_steps - step - 1, ref.shape[0], z.device)
                ref_noisy = self.q_sample(ref, ref_ts, ref_noise)
                image = ref_noisy * ref_mask + image * (1.0 - ref_mask)
        self.initialized = False
        return image


class UncondSamplerMixin:
    model: IDiffusion
    uncond: Optional[Tensor]
    unconditional_cond: Optional[Any]
    uncond_guidance_scale: float

    def _reset_uncond_buffers(
        self,
        unconditional_cond: Optional[Any],
        unconditional_guidance_scale: float,
    ) -> None:
        if unconditional_cond is None or self.model.condition_model is None:
            self.uncond = None
            self.uncond_guidance_scale = 0.0
        else:
            self.uncond = self.model._get_cond(unconditional_cond)
            self.uncond_guidance_scale = unconditional_guidance_scale

    def _uncond_denoise(
        self,
        image: Tensor,
        ts: Tensor,
        cond: Optional[cond_type],
        step: int,
        total_step: int,
    ) -> Tensor:
        if cond is None or self.uncond is None:
            return self.model.denoise(image, ts, cond, step, total_step)
        uncond = self.uncond.repeat_interleave(image.shape[0], dim=0)
        cond2 = None
        if not isinstance(cond, dict):
            if cond.shape[1] == uncond.shape[1]:
                cond2 = torch.cat([uncond, cond])
        else:
            cond2 = shallow_copy_dict(cond)
            for k, v in cond2.items():
                if is_misc_key(k):
                    continue
                if v.shape[1] != uncond.shape[1]:
                    cond2 = None
                    break
                cond2[k] = torch.cat([uncond, v])
        if cond2 is not None:
            image2 = torch.cat([image, image])
            ts2 = torch.cat([ts, ts])
            eps_uncond, eps = self.model.denoise(
                image2,
                ts2,
                cond2,
                step,
                total_step,
            ).chunk(2)
        else:
            eps = self.model.denoise(image, ts, cond, step, total_step)
            if not isinstance(cond, dict):
                uncond_cond = uncond
            else:
                uncond_cond = shallow_copy_dict(cond)
                for k, v in cond.items():
                    if not is_misc_key(k):
                        uncond_cond[k] = uncond
            eps_uncond = self.model.denoise(image, ts, uncond_cond, step, total_step)
        return eps_uncond + self.uncond_guidance_scale * (eps - eps_uncond)


__all__ = [
    "is_misc_key",
    "ISampler",
    "IQSampler",
    "DDPMQSampler",
]



from typing import Any
from typing import Optional
from cftool.array import l2_normalize
from cftool.types import tensor_dict_type

from .core import HuggingFaceModel
from ....schema import TrainerState
from ....constants import LATENT_KEY


@HuggingFaceModel.register("simbert")
class SimBERT(HuggingFaceModel):
    def __init__(self) -> None:
        super().__init__("peterchou/simbert-chinese-base")

    def forward(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: Optional["TrainerState"] = None,
        **kwargs: Any,
    ) -> tensor_dict_type:
        results = super().forward(batch_idx, batch, state, **kwargs)
        results[LATENT_KEY] = l2_normalize(results["pooler_output"])
        return results


__all__ = [
    "SimBERT",
]



import numpy as np
import torch

from .model import handpose_model
from . import util

try:
    import cv2
except:
    cv2 = None
try:
    from scipy.ndimage.filters import gaussian_filter
except:
    gaussian_filter = None
try:
    from skimage.measure import label
except:
    label = None


class Hand(object):
    def __init__(self, model_path, device="cpu"):
        if cv2 is None:
            raise ValueError("`cv2` is needed for `Body`")
        if gaussian_filter is None:
            raise ValueError("`scipy` is needed for `Body`")
        if label is None:
            raise ValueError("`scikit-image` is needed for `Hand`")
        self.model = handpose_model()
        model_dict = util.transfer(self.model, torch.load(model_path))
        self.model.load_state_dict(model_dict)
        self.model.eval()
        self.model.to(device)

    def __call__(self, oriImg, to):
        scale_search = [0.5, 1.0, 1.5, 2.0]
        # scale_search = [0.5]
        boxsize = 368
        stride = 8
        padValue = 128
        thre = 0.05
        multiplier = [x * boxsize / oriImg.shape[0] for x in scale_search]
        heatmap_avg = np.zeros((oriImg.shape[0], oriImg.shape[1], 22))
        # paf_avg = np.zeros((oriImg.shape[0], oriImg.shape[1], 38))

        for m in range(len(multiplier)):
            scale = multiplier[m]
            imageToTest = cv2.resize(
                oriImg, (0, 0), fx=scale, fy=scale, interpolation=cv2.INTER_CUBIC
            )
            imageToTest_padded, pad = util.padRightDownCorner(
                imageToTest, stride, padValue
            )
            im = (
                np.transpose(
                    np.float32(imageToTest_padded[:, :, :, np.newaxis]), (3, 2, 0, 1)
                )
                / 256
                - 0.5
            )
            im = np.ascontiguousarray(im)

            data = to(torch.from_numpy(im))
            # data = data.permute([2, 0, 1]).unsqueeze(0).float()
            with torch.no_grad():
                output = self.model(data).cpu().float().numpy()
                # output = self.model(data).numpy()q

            # extract outputs, resize, and remove padding
            heatmap = np.transpose(
                np.squeeze(output), (1, 2, 0)
            )  # output 1 is heatmaps
            heatmap = cv2.resize(
                heatmap, (0, 0), fx=stride, fy=stride, interpolation=cv2.INTER_CUBIC
            )
            heatmap = heatmap[
                : imageToTest_padded.shape[0] - pad[2],
                : imageToTest_padded.shape[1] - pad[3],
                :,
            ]
            heatmap = cv2.resize(
                heatmap,
                (oriImg.shape[1], oriImg.shape[0]),
                interpolation=cv2.INTER_CUBIC,
            )

            heatmap_avg += heatmap / len(multiplier)

        all_peaks = []
        for part in range(21):
            map_ori = heatmap_avg[:, :, part]
            one_heatmap = gaussian_filter(map_ori, sigma=3)
            binary = np.ascontiguousarray(one_heatmap > thre, dtype=np.uint8)
            # 全部小于阈值
            if np.sum(binary) == 0:
                all_peaks.append([0, 0])
                continue
            label_img, label_numbers = label(
                binary, return_num=True, connectivity=binary.ndim
            )
            max_index = (
                np.argmax(
                    [
                        np.sum(map_ori[label_img == i])
                        for i in range(1, label_numbers + 1)
                    ]
                )
                + 1
            )
            label_img[label_img != max_index] = 0
            map_ori[label_img == 0] = 0

            y, x = util.npmax(map_ori)
            all_peaks.append([x, y])
        return np.array(all_peaks)


if __name__ == "__main__":
    hand_estimation = Hand("../model/hand_pose_model.pth")

    # test_image = '../images/hand.jpg'
    test_image = "../images/hand.jpg"
    oriImg = cv2.imread(test_image)  # B,G,R order
    peaks = hand_estimation(oriImg)
    canvas = util.draw_handpose(oriImg, peaks, True)
    cv2.imshow("", canvas)
    cv2.waitKey(0)



from typing import Any
from typing import Union
from cftool.types import tensor_dict_type

from ...schema import DLConfig
from ...zoo.core import DLZoo
from ...zoo.core import TPipeline
from ...models.cv.diffusion.utils import CONCAT_TYPE


TLoadZoo = Union[TPipeline, DLConfig]


def _load(model: str, **kwargs: Any) -> TLoadZoo:
    fn = DLZoo.load_config if kwargs.pop("load_config", False) else DLZoo.load_pipeline
    return fn(model, **kwargs)  # type: ignore


# ae


def ae_kl_f4(size: int = 256, pretrained: bool = True, **kwargs: Any) -> TLoadZoo:
    if pretrained and size != 256:
        raise ValueError("pretrained `ae_kl_f4` should have `size`=256")
    kwargs["img_size"] = size
    kwargs["pretrained"] = pretrained
    return _load("ae/kl.f4", **kwargs)


def ae_kl_f8(size: int = 256, pretrained: bool = True, **kwargs: Any) -> TLoadZoo:
    if pretrained and size != 256:
        raise ValueError("pretrained `ae_kl_f8` should have `size`=256")
    kwargs["img_size"] = size
    kwargs["pretrained"] = pretrained
    return _load("ae/kl.f8", **kwargs)


def ae_kl_f16(size: int = 256, pretrained: bool = True, **kwargs: Any) -> TLoadZoo:
    if pretrained and size != 256:
        raise ValueError("pretrained `ae_kl_f16` should have `size`=256")
    kwargs["img_size"] = size
    kwargs["pretrained"] = pretrained
    return _load("ae/kl.f16", **kwargs)


# diffusion


def ddpm(img_size: int = 256, **kwargs: Any) -> TLoadZoo:
    kwargs["img_size"] = img_size
    return _load("diffusion/ddpm", **kwargs)


def _ldm(
    model: str,
    latent_size: int,
    latent_in_channels: int,
    latent_out_channels: int,
    **kwargs: Any,
) -> TLoadZoo:
    kwargs["img_size"] = latent_size
    kwargs["in_channels"] = latent_in_channels
    kwargs["out_channels"] = latent_out_channels
    model_config = kwargs.setdefault("model_config", {})
    first_stage_kw = model_config.setdefault("first_stage_config", {})
    first_stage_kw.setdefault("pretrained", False)
    first_stage_model_config = first_stage_kw.setdefault("model_config", {})
    use_loss = first_stage_model_config.setdefault("use_loss", False)
    if not use_loss:

        def state_callback(states: tensor_dict_type) -> tensor_dict_type:
            for key in list(states.keys()):
                if key.startswith("loss"):
                    states.pop(key)
            return states

        first_stage_kw["states_callback"] = state_callback
    return _load(model, **kwargs)


def ldm(
    latent_size: int = 32,
    latent_in_channels: int = 4,
    latent_out_channels: int = 4,
    **kwargs: Any,
) -> TLoadZoo:
    return _ldm(
        "diffusion/ldm",
        latent_size,
        latent_in_channels,
        latent_out_channels,
        **kwargs,
    )


def ldm_vq(
    latent_size: int = 64,
    latent_in_channels: int = 3,
    latent_out_channels: int = 3,
    **kwargs: Any,
) -> TLoadZoo:
    return _ldm(
        "diffusion/ldm.vq",
        latent_size,
        latent_in_channels,
        latent_out_channels,
        **kwargs,
    )


def ldm_sd(pretrained: bool = True, **kwargs: Any) -> TLoadZoo:
    return _ldm("diffusion/ldm.sd", 64, 4, 4, pretrained=pretrained, **kwargs)


def ldm_sd_tag(tag: str, pretrained: bool = True) -> TLoadZoo:
    return ldm_sd(pretrained, download_name=f"ldm_sd_{tag}")


def ldm_sd_inpainting(pretrained: bool = True, **kw: Any) -> TLoadZoo:
    return _ldm("diffusion/ldm.sd_inpainting", 64, 9, 4, pretrained=pretrained, **kw)


def ldm_sd_v2(pretrained: bool = True, **kwargs: Any) -> TLoadZoo:
    return _ldm("diffusion/ldm.sd_v2", 64, 4, 4, pretrained=pretrained, **kwargs)


def ldm_sd_v2_base(pretrained: bool = True, **kwargs: Any) -> TLoadZoo:
    return _ldm("diffusion/ldm.sd_v2_base", 64, 4, 4, pretrained=pretrained, **kwargs)


def ldm_celeba_hq(pretrained: bool = True) -> TLoadZoo:
    return ldm_vq(
        pretrained=pretrained,
        download_name="ldm_celeba_hq",
        model_config=dict(
            ema_decay=None,
            first_stage_config=dict(
                pretrained=False,
            ),
        ),
    )


def ldm_inpainting(pretrained: bool = True) -> TLoadZoo:
    return ldm_vq(
        pretrained=pretrained,
        latent_in_channels=7,
        download_name="ldm_inpainting",
        model_config=dict(
            ema_decay=None,
            start_channels=256,
            num_heads=8,
            num_head_channels=None,
            resample_with_resblock=True,
            condition_type=CONCAT_TYPE,
            first_stage_config=dict(
                pretrained=False,
                model_config=dict(
                    attention_type="none",
                ),
            ),
        ),
    )


def ldm_sr(pretrained: bool = True) -> TLoadZoo:
    return ldm_vq(
        pretrained=pretrained,
        latent_in_channels=6,
        download_name="ldm_sr",
        model_config=dict(
            ema_decay=None,
            start_channels=160,
            attention_downsample_rates=[8, 16],
            channel_multipliers=[1, 2, 2, 4],
            condition_type=CONCAT_TYPE,
            first_stage_config=dict(
                pretrained=False,
            ),
        ),
    )


def ldm_semantic(pretrained: bool = True) -> TLoadZoo:
    return ldm_vq(
        pretrained=pretrained,
        latent_size=128,
        latent_in_channels=6,
        download_name="ldm_semantic",
        model_config=dict(
            ema_decay=None,
            start_channels=128,
            num_heads=8,
            num_head_channels=None,
            attention_downsample_rates=[8, 16, 32],
            channel_multipliers=[1, 4, 8],
            condition_type=CONCAT_TYPE,
            condition_model="rescaler",
            condition_config=dict(
                num_stages=2,
                in_channels=182,
                out_channels=3,
            ),
            first_stage_config=dict(
                pretrained=False,
            ),
        ),
    )



import torch.nn as nn

from ..mconfigs import ALL_MCONFIGS
from ..utils.misc import load_weights


def load_model(model_type: str, ckpt_path: str, verbose: bool = False) -> nn.Module:
    net = ALL_MCONFIGS[model_type]["model"](**ALL_MCONFIGS[model_type]["params"])
    load_weights(net, ckpt_path, verbose=verbose)
    return net



import torch.nn as nn

from torch import Tensor
from typing import Any
from typing import List
from typing import Union
from typing import Optional
from cftool.misc import shallow_copy_dict
from cftool.types import tensor_dict_type

from .schema import DecoderMixin
from .schema import Decoder1DMixin
from ....constants import INPUT_KEY
from ....modules.blocks import get_conv_blocks
from ....modules.blocks import Lambda
from ....modules.blocks import Linear
from ....modules.blocks import Activation
from ....modules.blocks import ResidualBlock
from ....modules.blocks import ChannelPadding
from ....modules.blocks import UpsampleConv2d


@DecoderMixin.register("vanilla")
class VanillaDecoder(nn.Module, DecoderMixin):
    def __init__(
        self,
        latent_channels: int,
        out_channels: int,
        norm_type: Optional[str] = "instance",
        res_norm_type: Optional[str] = "instance",
        activation: Optional[str] = "leaky_relu_0.2",
        padding: str = "reflection",
        *,
        kernel_size: int = 3,
        last_kernel_size: Optional[int] = None,
        num_residual_blocks: int = 0,
        residual_dropout: float = 0.0,
        num_repeats: Union[str, List[int]] = "default",
        reduce_channel_on_upsample: bool = False,
        img_size: Optional[int] = None,
        num_upsample: Optional[int] = None,
        cond_channels: int = 16,
        num_classes: Optional[int] = None,
        latent_resolution: Optional[int] = None,
        final_activation: Optional[str] = None,
    ):
        super().__init__()
        self.latent_channels = latent_channels
        self._initialize(
            out_channels=out_channels,
            img_size=img_size,
            num_upsample=num_upsample,
            num_classes=num_classes,
            latent_resolution=latent_resolution,
        )
        self._init_cond(cond_channels=cond_channels)

        blocks: List[nn.Module] = []

        for _ in range(num_residual_blocks):
            blocks.append(
                ResidualBlock(
                    latent_channels,
                    residual_dropout,
                    norm_type=res_norm_type,
                    activation=activation,
                    padding=padding,
                )
            )

        if isinstance(num_repeats, str):
            if num_repeats == "default":
                num_repeats = [0] + [1] * self.num_upsample
            elif num_repeats == "repeated":
                repeats1 = (self.num_upsample - 1) // 2
                repeats0 = self.num_upsample - repeats1 - 1
                num_repeats = [1] + [4] * repeats0 + [2] * repeats1 + [1]
            else:
                raise ValueError(f"unrecognized `num_repeats` '{num_repeats}` occurred")
        if len(num_repeats) != self.num_upsample + 1:
            msg = "length of `num_repeats` is not identical with `num_upsample + 1`"
            raise ValueError(msg)

        if last_kernel_size is None:
            last_kernel_size = kernel_size
        in_nc = latent_channels
        for i, num_repeat in enumerate(num_repeats):
            is_last = i == self.num_upsample
            if is_last:
                num_repeat += 1
            if num_repeat == 0:
                continue
            repeat_channels = latent_channels if i == 0 else in_nc
            out_nc = repeat_channels // 2
            if i != 0:
                num_repeat -= 1
                if reduce_channel_on_upsample:
                    first_out_nc = out_nc
                else:
                    first_out_nc = in_nc if num_repeat != 0 else out_nc
                if is_last and num_repeat == 0:
                    first_out_nc = out_channels
                    kernel_size = last_kernel_size
                    norm_type = None
                    activation = None
                blocks.extend(
                    get_conv_blocks(
                        in_nc,
                        first_out_nc,
                        kernel_size,
                        1,
                        bias=True,
                        factor=2,
                        norm_type=norm_type,
                        activation=activation,
                        conv_base=UpsampleConv2d,
                        padding=padding,
                    )
                )
                in_nc = first_out_nc
            for j in range(num_repeat):
                if is_last and j == num_repeat - 1:
                    out_nc = out_channels
                    kernel_size = last_kernel_size
                    norm_type = None
                    activation = None
                blocks.extend(
                    get_conv_blocks(
                        in_nc,
                        repeat_channels if j != num_repeat - 1 else out_nc,
                        kernel_size,
                        1,
                        bias=True,
                        norm_type=norm_type,
                        activation=activation,
                        padding=padding,
                    )
                )
                in_nc = repeat_channels
            in_nc = out_nc

        self.decoder = nn.ModuleList(blocks)
        if final_activation is None:
            self.final_activation = None
        else:
            self.final_activation = Activation.make(final_activation)

    def forward(self, batch: tensor_dict_type, **kwargs: Any) -> Tensor:
        net = self._inject_cond(batch)[INPUT_KEY]
        determinate = kwargs.pop("determinate", False)
        for block in self.decoder:
            kw = {}
            if isinstance(block, UpsampleConv2d):
                kw["determinate"] = determinate
            net = block(net, **kw)
        net = self.resize(net, determinate=determinate)
        if self.final_activation is not None:
            net = self.final_activation(net)
        return net


@Decoder1DMixin.register("vanilla")
class VanillaDecoder1D(nn.Module, Decoder1DMixin):
    def __init__(
        self,
        latent_dim: int,
        out_channels: int,
        latent_channels: int = 64,
        norm_type: Optional[str] = "instance",
        res_norm_type: Optional[str] = "instance",
        activation: Optional[str] = "leaky_relu_0.2",
        padding: str = "reflection",
        *,
        kernel_size: int = 3,
        last_kernel_size: Optional[int] = None,
        num_residual_blocks: int = 0,
        residual_dropout: float = 0.0,
        num_repeats: Union[str, List[int]] = "default",
        reduce_channel_on_upsample: bool = False,
        img_size: Optional[int] = None,
        num_upsample: Optional[int] = None,
        cond_channels: int = 16,
        num_classes: Optional[int] = None,
        latent_resolution: Optional[int] = None,
        latent_padding_channels: Optional[int] = None,
        final_activation: Optional[str] = None,
    ):
        super().__init__()
        self.latent_dim = latent_dim
        self._initialize(
            out_channels=out_channels,
            img_size=img_size,
            num_upsample=num_upsample,
            num_classes=num_classes,
            latent_resolution=latent_resolution,
        )
        if latent_resolution is None:
            latent_resolution = int(round(img_size / 2**self.num_upsample))
        self.latent_resolution = latent_resolution
        assert isinstance(self.latent_resolution, int)
        latent_area = self.latent_resolution**2
        shape = -1, latent_channels, latent_resolution, latent_resolution
        blocks: List[nn.Module] = [
            Linear(latent_dim, latent_channels * latent_area),
            Activation.make(activation),
            Lambda(lambda tensor: tensor.view(*shape), f"reshape -> {shape}"),
        ]
        if latent_padding_channels is not None:
            latent_padding = ChannelPadding(
                latent_channels,
                latent_padding_channels,
                latent_resolution,
            )
            blocks.append(latent_padding)
        self.from_latent = nn.Sequential(*blocks)
        self.decoder = VanillaDecoder(
            latent_channels,
            out_channels,
            norm_type,
            res_norm_type,
            activation,
            padding,
            kernel_size=kernel_size,
            last_kernel_size=last_kernel_size,
            num_residual_blocks=num_residual_blocks,
            residual_dropout=residual_dropout,
            num_repeats=num_repeats,
            reduce_channel_on_upsample=reduce_channel_on_upsample,
            img_size=img_size,
            num_upsample=self.num_upsample,
            cond_channels=cond_channels,
            num_classes=num_classes,
            latent_resolution=latent_resolution,
            final_activation=final_activation,
        )

    def forward(self, batch: tensor_dict_type, **kwargs: Any) -> Tensor:
        batch = shallow_copy_dict(batch)
        net = batch[INPUT_KEY]
        net = self.from_latent(net)
        batch[INPUT_KEY] = net
        return self.decoder(batch, **kwargs)


__all__ = [
    "VanillaDecoder",
    "VanillaDecoder1D",
]



import os
import json
import torch

from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import Tuple
from typing import Union
from typing import Optional
from typing import NamedTuple
from cftool.misc import update_dict
from cftool.misc import safe_execute
from cftool.misc import shallow_copy_dict
from cftool.types import tensor_dict_type

from .models import dl_zoo_model_loaders
from ..schema import IData
from ..schema import DLConfig
from ..pipeline import Pipeline
from ..pipeline import PipelineTypes
from ..pipeline import TrainingPipeline
from ..pipeline import DLInferencePipeline
from ..constants import DEFAULT_ZOO_TAG
from ..parameters import OPT
from ..misc.toolkit import download_model


root = os.path.dirname(__file__)
configs_root = os.path.join(root, "configs")
TPipeline = Union[TrainingPipeline, DLInferencePipeline]


class ParsedModel(NamedTuple):
    json_path: str
    download_name: str


def _parse_model(model: str) -> ParsedModel:
    tag = DEFAULT_ZOO_TAG
    model_type, model_name = model.split("/")
    download_name = model_name
    if "." in model_name:
        model_name, tag = model_name.split(".")
    json_folder = os.path.join(configs_root, model_type, model_name)
    json_path = os.path.join(json_folder, f"{tag}.json")
    if not os.path.isfile(json_path):
        json_path = os.path.join(json_folder, f"{DEFAULT_ZOO_TAG}.json")
    return ParsedModel(json_path, download_name)


def _parse_config(json_path: str) -> Dict[str, Any]:
    with open(json_path, "r") as f:
        config = json.load(f)
    bases = config.pop("__bases__", [])
    for base in bases[::-1]:
        parsed = _parse_model(base)
        config = update_dict(config, _parse_config(parsed.json_path))
    return config


class DLZoo:
    model_dir = "models_v0.4.x"

    # api

    @classmethod
    def load_config(
        cls,
        model: Optional[str] = None,
        *,
        json_path: Optional[str] = None,
        download_name: Optional[str] = None,
        **kwargs: Any,
    ) -> DLConfig:
        return cls._load_config(
            model,
            json_path=json_path,
            download_name=download_name,
            **kwargs,
        )[0]

    @classmethod
    def load_pipeline(
        cls,
        model: Optional[str] = None,
        *,
        data: Optional[IData] = None,
        states: Optional[tensor_dict_type] = None,
        pretrained: bool = False,
        pipeline_type: PipelineTypes = PipelineTypes.DL_INFERENCE,
        download_name: Optional[str] = None,
        model_dir: Optional[str] = None,
        json_path: Optional[str] = None,
        **kwargs: Any,
    ) -> TPipeline:
        states_callback = kwargs.pop("states_callback", None)
        # get config
        config, download_name = cls._load_config(
            model,
            json_path=json_path,
            download_name=download_name,
            **kwargs,
        )
        # handle states
        if states is None and pretrained:
            if download_name is None:
                err_msg = f"`{'tag'}` should be provided in '{json_path}' when `pretrained` is True"
                raise ValueError(err_msg)
            if model_dir is None:
                model_dir = cls.model_dir
            root = os.path.join(OPT.cache_dir, model_dir)
            states_path = download_model(download_name, root=root)
            states = torch.load(states_path, map_location="cpu")
        if states is not None and states_callback is not None:
            states = states_callback(states)
        # build
        m_base: Type[Pipeline] = Pipeline.get(pipeline_type)
        if issubclass(m_base, DLInferencePipeline):
            return m_base.build_with(config, states)
        m = m_base.init(config)
        if isinstance(m, TrainingPipeline):
            if states is not None:
                if data is None:
                    msg = "`data` needs to be provided when loading `TrainingPipeline` with `states`"
                    raise ValueError(msg)
                m.prepare(data)
                m.build_model.model.load_state_dict(states)
        return m

    # internal

    @classmethod
    def _load_config(
        cls,
        model: Optional[str] = None,
        *,
        json_path: Optional[str] = None,
        download_name: Optional[str] = None,
        **kwargs: Any,
    ) -> Tuple[DLConfig, Optional[str]]:
        # loader
        if model is not None:
            loader_base = dl_zoo_model_loaders.get(model)
            if loader_base is not None:
                loader_base().permute_kwargs(kwargs)
        # load json
        if json_path is None:
            if model is None:
                raise ValueError("either `model` or `json_path` should be provided")
            parsed = _parse_model(model)
            json_path = parsed.json_path
            if download_name is None:
                download_name = parsed.download_name
        parsed_config = _parse_config(json_path)
        if download_name is None:
            download_name = parsed_config.pop("tag", None)
        # handle requires

        def _example(l: str, r: str, h: List[str]) -> str:
            key = h.pop(0)
            if not h:
                return f"{l}{key}=...{r}"
            return _example(f"{l}{key}=dict(", f"){r}", h)

        def _inject_requires(
            inc: Dict[str, Any],
            reference: Dict[str, Any],
            local_requires: Dict[str, Any],
            hierarchy: Optional[str],
        ) -> None:
            for k, v in local_requires.items():
                k_hierarchy = k if hierarchy is None else f"{hierarchy} -> {k}"
                ki = inc.setdefault(k, {})
                kr = reference.setdefault(k, {})
                if isinstance(v, dict):
                    _inject_requires(ki, kr, v, k_hierarchy)
                    continue
                assert isinstance(v, list), "requirements should be a list"
                for vv in v:
                    shortcut = kwargs.pop(vv, no_shortcut_token)
                    if shortcut != no_shortcut_token:
                        ki[vv] = shortcut
                        continue
                    if vv not in kr:
                        example = _example("", "", k_hierarchy.split(" -> ") + [vv])
                        raise ValueError(
                            f"Failed to build '{model}': "
                            f"'{vv}' should be provided in `{k_hierarchy}`, for example:\n"
                            f'* cflearn.api.from_zoo("{model}", {example})\n'
                            f'* cflearn.ZooBase.load_pipeline("{model}", {example})\n'
                            f"\nOr you can use shortcut for simplicity:\n"
                            f'* cflearn.api.from_zoo("{model}", {vv}=...)\n'
                            f'* cflearn.ZooBase.load_pipeline("{model}", {vv}=...)\n'
                        )

        no_shortcut_token = "@#$no_shortcut$#@"
        requires = parsed_config.pop("__requires__", {})
        merged_config = shallow_copy_dict(parsed_config)
        update_dict(shallow_copy_dict(kwargs), merged_config)
        increment: Dict[str, Any] = {}
        _inject_requires(increment, merged_config, requires, None)
        raw_config = shallow_copy_dict(parsed_config)
        update_dict(kwargs, raw_config)
        update_dict(increment, raw_config)
        config_type = raw_config.pop("config_type", "dl")
        config = safe_execute(DLConfig.get(config_type), raw_config)
        return config, download_name


__all__ = [
    "DLZoo",
]



import torch

import numpy as np

from torch import nn
from typing import Tuple

from .transforms import ToTensor
from .transforms import PadToDivisor
from .transforms import NormalizeTensor
from .transforms import AddFlippedTensor
from ..model.backboned import HRNetIHModel


class Predictor:
    def __init__(
        self,
        net: nn.Module,
        device: torch.device,
        *,
        with_flip: bool = False,
        mean: Tuple[float, float, float] = (0.485, 0.456, 0.406),
        std: Tuple[float, float, float] = (0.229, 0.224, 0.225),
    ):
        self.device = device
        self.net = net.to(self.device)
        self.net.eval()

        if hasattr(net, "depth"):
            size_divisor = 2 ** (net.depth + 1)
        elif isinstance(net, HRNetIHModel):
            size_divisor = 128
        else:
            size_divisor = 1

        mean = torch.tensor(mean, dtype=torch.float32)
        std = torch.tensor(std, dtype=torch.float32)
        self.transforms = [
            PadToDivisor(divisor=size_divisor, border_mode=0),
            ToTensor("cpu"),
            NormalizeTensor(mean, std, "cpu"),
        ]
        if with_flip:
            self.transforms.append(AddFlippedTensor())

    def to(self, device: torch.device) -> None:
        self.device = device
        self.net.to(device)

    # return uint8 image array
    @torch.no_grad()
    def predict(self, image: np.ndarray, mask: np.ndarray) -> np.ndarray:
        for transform in self.transforms:
            image, mask = transform.transform(image, mask)
        image, mask = image.to(self.device), mask.to(self.device)
        predicted_image = self.net(image, mask)["images"].cpu()
        for transform in reversed(self.transforms):
            predicted_image = transform.inv_transform(predicted_image)
        predicted_image = torch.clamp(predicted_image, 0, 255).to(torch.uint8)
        return predicted_image.cpu().numpy()



import torch

import torch.nn.functional as F

from PIL import Image
from torch import Tensor
from typing import Tuple
from typing import Union
from typing import Optional
from cftool.cv import read_image
from cftool.cv import save_images

from ..utils import APIMixin
from ...zoo import DLZoo
from ...zoo.core import TPipeline
from ...misc.toolkit import eval_context


class TranslatorAPI(APIMixin):
    def sr(
        self,
        image: Union[str, Image.Image],
        export_path: Optional[str] = None,
        *,
        max_wh: int = 768,
        clip_range: Optional[Tuple[int, int]] = (0, 1),
    ) -> Tensor:
        res = read_image(image, max_wh, anchor=None)
        # inference
        tensor = torch.from_numpy(res.image)
        tensor = tensor.contiguous().to(self.device)
        if self.use_half:
            tensor = tensor.half()
        with eval_context(self.m):
            output = self.m(tensor).cpu().float()
        if clip_range is not None:
            output = torch.clip(output, *clip_range)
        # handle alpha
        if res.alpha is not None:
            alpha_tensor = torch.from_numpy(res.alpha)
            with torch.no_grad():
                alpha_tensor = F.interpolate(
                    alpha_tensor,
                    output.shape[-2:],
                    mode="nearest",
                )
            output = torch.cat([output, alpha_tensor], dim=1)
        # export
        if export_path is not None:
            save_images(output, export_path)
        self.empty_cuda_cache()
        return output

    @classmethod
    def from_esr(
        cls,
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> "TranslatorAPI":
        return cls.from_pipeline(esr(), device, use_amp=use_amp, use_half=use_half)

    @classmethod
    def from_esr_anime(
        cls,
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> "TranslatorAPI":
        m = esr_anime()
        return cls.from_pipeline(m, device, use_amp=use_amp, use_half=use_half)


def esr(pretrained: bool = True) -> TPipeline:
    return DLZoo.load_pipeline("sr/esr", pretrained=pretrained)


def esr_anime(pretrained: bool = True) -> TPipeline:
    return DLZoo.load_pipeline("sr/esr.anime", pretrained=pretrained)


__all__ = [
    "TranslatorAPI",
]



import argparse

from typing import Any
from typing import Dict
from typing import Optional
from typing import NamedTuple
from cftool.misc import Serializer

from cflearn.data import MLData
from cflearn.schema import DLConfig
from cflearn.dist.ml.task import Task


class Info(NamedTuple):
    workplace: str
    meta: Dict[str, Any]
    config: Optional[DLConfig]
    data: Optional[MLData]


def get_info(*, requires_data: bool = True) -> Info:
    # parse
    parser = argparse.ArgumentParser()
    parser.add_argument("--task_folder", type=str)
    args = parser.parse_args()
    task_folder = args.task_folder
    # common
    task = Serializer.load(task_folder, Task)
    meta_config = task.meta_kwargs
    workspace = meta_config["workspace"]
    if task.config is None:
        config = None
    else:
        config = DLConfig.from_pack(task.config)
    # data
    if not requires_data:
        data = None
    else:
        data_folder = meta_config.pop("data_folder", None)
        if data_folder is None:
            msg = "`data_folder` should be provided when `requires_data` is True"
            raise ValueError(msg)
        data = Serializer.load(data_folder, MLData)
    return Info(workspace, meta_config, config, data)


__all__ = [
    "get_info",
]



import torch

import numpy as np
import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union
from typing import Callable
from typing import Optional
from cftool.misc import safe_execute
from cftool.misc import shallow_copy_dict
from cftool.array import to_torch
from cftool.types import tensor_dict_type

from .unet import ControlNet
from .unet import UNetDiffuser
from .utils import cond_type
from .utils import extract_to
from .utils import get_timesteps
from .utils import ADM_KEY
from .utils import ADM_TYPE
from .utils import CONCAT_KEY
from .utils import CONCAT_TYPE
from .utils import HYBRID_TYPE
from .utils import CROSS_ATTN_KEY
from .utils import CROSS_ATTN_TYPE
from .utils import CONTROL_HINT_KEY
from .utils import CONTROL_HINT_START_KEY
from .samplers import is_misc_key
from .samplers import ISampler
from .samplers import DDPMQSampler
from .cond_models import condition_models
from .cond_models import specialized_condition_models
from ...schemas import CustomTrainStep
from ...schemas import CustomTrainStepLoss
from ...schemas import ModelWithCustomSteps
from ...schemas import GaussianGeneratorMixin
from ....zoo import DLZoo
from ....schema import ITrainer
from ....schema import TrainerState
from ....schema import MetricsOutputs
from ....constants import INPUT_KEY
from ....constants import PREDICTIONS_KEY
from ....misc.toolkit import freeze
from ....misc.toolkit import get_dtype
from ....misc.toolkit import get_device
from ....modules.blocks import EMA


def make_beta_schedule(
    schedule: str,
    timesteps: int,
    linear_start: float,
    linear_end: float,
    cosine_s: float,
) -> np.ndarray:
    if schedule == "linear":
        betas = (
            np.linspace(
                linear_start**0.5,
                linear_end**0.5,
                timesteps,
                dtype=np.float64,
            )
            ** 2
        )
    elif schedule == "cosine":
        arange = np.arange(timesteps + 1, dtype=np.float64)
        timesteps = arange / timesteps + cosine_s
        alphas = timesteps / (1 + cosine_s) * np.pi / 2
        alphas = np.cos(alphas) ** 2
        alphas = alphas / alphas[0]
        betas = 1.0 - alphas[1:] / alphas[:-1]
        betas = np.clip(betas, 0.0, 0.999)
    else:
        lin = np.linspace(linear_start, linear_end, timesteps, dtype=np.float64)
        if schedule == "sqrt_linear":
            betas = lin
        elif schedule == "sqrt":
            betas = lin**0.5
        else:
            raise ValueError(f"unrecognized schedule '{schedule}' occurred")
    return betas


def make_condition_model(key: str, m: nn.Module) -> nn.Module:
    if key in specialized_condition_models:
        return m
    condition_base = condition_models.get(key)
    if condition_base is None:
        return m
    return condition_base(m)


class DDPMStep(CustomTrainStep):
    def loss_fn(
        self,
        m: "DDPM",
        state: Optional[TrainerState],
        batch: tensor_dict_type,
        forward_results: tensor_dict_type,
        **kwargs: Any,
    ) -> CustomTrainStepLoss:
        noise = forward_results[m.noise_key]
        unet_out = forward_results[PREDICTIONS_KEY]
        timesteps = forward_results[m.timesteps_key]
        if m.parameterization == "eps":
            target = noise
        elif m.parameterization == "x0":
            target = batch[INPUT_KEY]
        elif m.parameterization == "v":
            target = self.get_v(m, batch[INPUT_KEY], noise, timesteps)
        else:
            msg = f"unrecognized parameterization '{m.parameterization}' occurred"
            raise ValueError(msg)

        losses = {}
        if m.loss_type == "l1":
            loss = (unet_out - target).abs()
        elif m.loss_type == "l2":
            loss = F.mse_loss(unet_out, target, reduction="none")
        else:
            raise ValueError(f"unrecognized loss '{m.loss_type}' occurred")
        loss = loss.mean(dim=(1, 2, 3))
        loss_simple = loss
        losses["simple"] = loss_simple.mean().item()

        log_var_t = m.log_var[timesteps].to(unet_out.device)  # type: ignore
        loss_simple = loss_simple / torch.exp(log_var_t) + log_var_t
        if m.learn_log_var:
            losses["gamma"] = loss_simple.mean().item()
            losses["log_var"] = m.log_var.data.mean().item()  # type: ignore

        loss_simple = m.l_simple_weight * loss_simple.mean()
        if m.original_elbo_weight <= 0:
            losses["loss"] = loss_simple.item()
            return CustomTrainStepLoss(loss_simple, losses)

        loss_vlb = (m.lvlb_weights[timesteps] * loss).mean()
        losses["vlb"] = loss_vlb.item()

        loss_vlb = m.original_elbo_weight * loss_vlb
        loss = loss_simple + loss_vlb
        losses["loss"] = loss.item()
        return CustomTrainStepLoss(loss, losses)

    def get_v(self, m: "DDPM", x: Tensor, noise: Tensor, ts: Tensor) -> Tensor:
        num_dim = x.ndim
        return (
            extract_to(m.sqrt_alphas_cumprod, ts, num_dim) * noise
            - extract_to(m.sqrt_one_minus_alphas_cumprod, ts, num_dim) * x
        )

    def callback(
        self,
        m: "DDPM",
        trainer: ITrainer,
        batch: tensor_dict_type,
        forward_results: Union[tensor_dict_type, List[tensor_dict_type]],
    ) -> None:
        if m.training and m.unet_ema is not None:
            m.unet_ema()


@ModelWithCustomSteps.register("ddpm")
class DDPM(ModelWithCustomSteps, GaussianGeneratorMixin):
    cond_key = "cond"
    noise_key = "noise"
    timesteps_key = "timesteps"
    identity_condition_model = "identity"

    sampler: ISampler
    control_model: Optional[Union[ControlNet, nn.ModuleDict]]
    control_scales: Optional[Union[List[float], List[List[float]]]]

    def __init__(
        self,
        img_size: int,
        # unet
        in_channels: int,
        out_channels: int,
        *,
        start_channels: int = 320,
        num_heads: Optional[int] = 8,
        num_head_channels: Optional[int] = None,
        use_spatial_transformer: bool = False,
        num_transformer_layers: int = 1,
        context_dim: Optional[int] = None,
        signal_dim: int = 2,
        num_res_blocks: int = 2,
        attention_downsample_rates: Tuple[int, ...] = (1, 2, 4),
        dropout: float = 0.0,
        channel_multipliers: Tuple[int, ...] = (1, 2, 4, 4),
        resample_with_conv: bool = True,
        resample_with_resblock: bool = False,
        use_scale_shift_norm: bool = False,
        num_classes: Optional[int] = None,
        use_linear_in_transformer: bool = False,
        use_checkpoint: bool = False,
        attn_split_chunk: Optional[int] = None,
        tome_info: Optional[Dict[str, Any]] = None,
        # ControlNet
        only_mid_control: bool = False,
        # diffusion
        ema_decay: Optional[float] = None,
        use_num_updates_in_ema: bool = True,
        parameterization: str = "eps",
        ## condition
        condition_type: str = CROSS_ATTN_TYPE,
        condition_model: Optional[str] = None,
        condition_config: Optional[Dict[str, Any]] = None,
        condition_learnable: bool = False,
        use_pretrained_condition: bool = False,
        ## noise schedule
        v_posterior: float = 0.0,
        timesteps: int = 1000,
        given_betas: Optional[np.ndarray] = None,
        beta_schedule: str = "linear",
        linear_start: float = 1.0e-4,
        linear_end: float = 2.0e-2,
        cosine_s: float = 8.0e-3,
        ## loss
        loss_type: str = "l2",
        l_simple_weight: float = 1.0,
        original_elbo_weight: float = 0.0,
        learn_log_var: bool = False,
        log_var_init: float = 0.0,
        ## sampling
        sampler: str = "ddim",
        sampler_config: Optional[Dict[str, Any]] = None,
    ):
        super().__init__()
        self.img_size = img_size
        # unet
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.unet_kw = dict(
            in_channels=in_channels,
            context_dim=context_dim,
            num_heads=num_heads,
            num_head_channels=num_head_channels,
            use_spatial_transformer=use_spatial_transformer,
            num_transformer_layers=num_transformer_layers,
            signal_dim=signal_dim,
            start_channels=start_channels,
            num_res_blocks=num_res_blocks,
            attention_downsample_rates=attention_downsample_rates,
            dropout=dropout,
            channel_multipliers=channel_multipliers,
            resample_with_conv=resample_with_conv,
            resample_with_resblock=resample_with_resblock,
            use_scale_shift_norm=use_scale_shift_norm,
            num_classes=num_classes,
            use_linear_in_transformer=use_linear_in_transformer,
            use_checkpoint=use_checkpoint,
            attn_split_chunk=attn_split_chunk,
            tome_info=tome_info,
        )
        self.unet = UNetDiffuser(out_channels=out_channels, **self.unet_kw)  # type: ignore
        # ControlNet
        self.control_model = None
        self.only_mid_control = only_mid_control
        self.num_control_scales = len(self.unet.output_blocks) + 1
        self.control_scales = None
        # ema
        if ema_decay is None:
            self.unet_ema = None
        else:
            self.unet_ema = EMA(
                ema_decay,
                list(self.unet.named_parameters()),
                use_num_updates=use_num_updates_in_ema,
            )
        # condition
        self.condition_type = condition_type
        self.condition_learnable = condition_learnable
        self.use_pretrained_condition = use_pretrained_condition
        self._initialize_condition_model(
            condition_model,
            condition_config,
            condition_learnable,
            use_pretrained_condition,
        )
        # settings
        self.parameterization = parameterization
        self.v_posterior = v_posterior
        # noise schedule
        self._register_noise_schedule(
            timesteps,
            given_betas,
            beta_schedule,
            linear_start,
            linear_end,
            cosine_s,
        )
        # loss
        self.loss_type = loss_type
        self.l_simple_weight = l_simple_weight
        self.original_elbo_weight = original_elbo_weight
        self.learn_log_var = learn_log_var
        log_var = torch.full(fill_value=log_var_init, size=(self.t,))
        if not learn_log_var:
            self.log_var = log_var
        else:
            self.log_var = nn.Parameter(log_var, requires_grad=True)
        # sampler
        self.switch_sampler(sampler, sampler_config)
        self.q_sampler = DDPMQSampler(self)
        self.q_sampler.reset_buffers(
            self.sqrt_alphas_cumprod,
            self.sqrt_one_minus_alphas_cumprod,
        )

    @property
    def can_reconstruct(self) -> bool:
        return True

    @property
    def learnable(self) -> List[nn.Parameter]:
        params = list(self.unet.parameters())  # type: ignore
        if self.learn_log_var:
            params.append(self.log_var)
        return params

    @property
    def train_steps(self) -> List[CustomTrainStep]:
        return [DDPMStep("learnable")]

    def get_forward_args(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: Optional["TrainerState"] = None,
        **kwargs: Any,
    ) -> Tuple[Any, ...]:
        return batch[INPUT_KEY], batch.get(self.cond_key)

    def forward(self, net: Tensor, cond: Optional[Tensor]) -> tensor_dict_type:
        timesteps = torch.randint(0, self.t, (net.shape[0],), device=net.device).long()
        net = self._preprocess(net)
        noise = torch.randn_like(net)
        net = self._q_sample(net, timesteps, noise)
        unet_out = self.denoise(net, timesteps, cond, timesteps[0].item(), self.t)
        return {
            PREDICTIONS_KEY: unet_out,
            self.noise_key: noise,
            self.timesteps_key: timesteps,
        }

    def evaluate(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: Optional[TrainerState],
        weighted_loss_score_fn: Callable[[Dict[str, float]], float],
        forward_kwargs: Dict[str, Any],
    ) -> MetricsOutputs:
        train_step = self.train_steps[0]
        # TODO : specify timesteps & noise to make results deterministic
        forward = self.run(batch_idx, batch, state, **forward_kwargs)
        losses = train_step.loss_fn(self, state, batch, forward).losses
        score = -losses["simple"]
        # no ema
        if self.unet_ema is None:
            return MetricsOutputs(score, losses, {k: False for k in losses})
        losses = {f"{k}_ema": v for k, v in losses.items()}
        self.unet_ema.train()
        forward = self.run(batch_idx, batch, state, **forward_kwargs)
        losses.update(train_step.loss_fn(self, state, batch, forward).losses)
        self.unet_ema.eval()
        return MetricsOutputs(score, losses, {k: False for k in losses})

    # api

    def make_sampler(
        self,
        sampler: str,
        sampler_config: Optional[Dict[str, Any]] = None,
    ) -> ISampler:
        kw = shallow_copy_dict(sampler_config or {})
        kw["model"] = self
        return ISampler.make(sampler, kw)

    def switch_sampler(
        self,
        sampler: str,
        sampler_config: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.sampler = self.make_sampler(sampler, sampler_config)

    def generate_z(self, num_samples: int) -> Tensor:
        shape = num_samples, self.in_channels, self.img_size, self.img_size
        return torch.randn(shape, dtype=get_dtype(self), device=get_device(self))

    def decode(
        self,
        z: Tensor,
        *,
        cond: Optional[Any] = None,
        num_steps: Optional[int] = None,
        start_step: Optional[int] = None,
        verbose: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        return self.sampler.sample(
            z,
            cond=cond,
            num_steps=num_steps,
            start_step=start_step,
            verbose=verbose,
            **kwargs,
        )

    def sample(
        self,
        num_samples: int,
        *,
        cond: Optional[Any] = None,
        num_steps: Optional[int] = None,
        start_step: Optional[int] = None,
        clip_output: bool = True,
        verbose: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        sampled = super().sample(
            num_samples,
            cond=cond,
            num_steps=num_steps,
            start_step=start_step,
            verbose=verbose,
            **kwargs,
        )
        if clip_output:
            sampled = torch.clip(sampled, -1.0, 1.0)
        return sampled

    def reconstruct(
        self,
        net: Tensor,
        *,
        noise_steps: Optional[int] = None,
        cond: Optional[Any] = None,
        **kwargs: Any,
    ) -> Tensor:
        net = self._preprocess(net)
        if noise_steps is None:
            noise_steps = self.t
        ts = get_timesteps(noise_steps - 1, net.shape[0], net.device)
        z = self._q_sample(net, ts)
        kw = shallow_copy_dict(kwargs)
        kw.update(dict(z=z, cond=cond))
        net = safe_execute(self.decode, kw)
        return net

    def denoise(
        self,
        image: Tensor,
        timesteps: Tensor,
        cond: Optional[cond_type],
        step: int,
        total_step: int,
    ) -> Tensor:
        net = image
        cond_kw = {}
        cond_type = self.condition_type
        if cond is not None and cond_type != "none":
            if cond_type == CONCAT_TYPE:
                if isinstance(cond, dict):
                    cond = cond[CONCAT_KEY]
                net = torch.cat([net, cond], dim=1)
            elif cond_type == CROSS_ATTN_TYPE:
                if not isinstance(cond, dict):
                    cond_kw = {CROSS_ATTN_KEY: cond}
                else:
                    cond_kw = {CROSS_ATTN_KEY: cond[CROSS_ATTN_KEY]}
            elif cond_type == HYBRID_TYPE:
                if not isinstance(cond, dict):
                    raise ValueError("`cond` should be a dict when `hybrid` is applied")
                concat = cond[CONCAT_KEY].repeat_interleave(len(net), dim=0)
                net = torch.cat([net, concat], dim=1)
                cond_kw = {CROSS_ATTN_KEY: cond[CROSS_ATTN_KEY]}
            elif cond_type == ADM_TYPE:
                if isinstance(cond, dict):
                    cond_kw = cond
                else:
                    cond_kw = {ADM_KEY: cond}
            else:
                raise ValueError(f"unrecognized condition type {cond_type} occurred")
        if self.control_model is not None:
            if not isinstance(cond, dict):
                raise ValueError("`cond` should be a dict when `control_model` is used")
            hint = cond.get(CONTROL_HINT_KEY)
            hint_start = cond.get(CONTROL_HINT_START_KEY)
            check_hint_start = lambda start: start is None or start * total_step <= step
            if hint is None:
                raise ValueError("`hint` should be provided for `control_model`")
            if isinstance(self.control_model, ControlNet):
                if not isinstance(hint, Tensor):
                    raise ValueError("`hint` should be a Tensor for single control")
                if hint_start is not None and isinstance(hint_start, dict):
                    raise ValueError("`hint_start` should be float for single control")
                if not check_hint_start(hint_start):
                    ctrl = None
                else:
                    ctrl = self.control_model(net, hint, timesteps=timesteps, **cond_kw)
                    if self.control_scales is None:
                        scales = [1.0] * self.num_control_scales
                    else:
                        if isinstance(self.control_scales[0], list):
                            raise ValueError("`control_scales` should be list of float")
                        scales = self.control_scales  # type: ignore
                    ctrl = [c * scale for c, scale in zip(ctrl, scales)]
            else:
                if not isinstance(hint, list):
                    raise ValueError("`hint` should be a list for control settings")
                if not isinstance(hint_start, list):
                    raise ValueError("`hint_start` should be a list of Optional[float]")
                target_keys = set(self.control_model.keys())
                hint_types = set(pair[0] for pair in hint)
                if hint_types - target_keys:
                    msg = f"`hint` ({hint_types}) should not exceed following keys: {', '.join(sorted(target_keys))}"
                    raise ValueError(msg)
                ctrl = [0.0] * self.num_control_scales
                any_activated = False
                for i, ((i_type, i_hint), i_start) in enumerate(zip(hint, hint_start)):
                    if not check_hint_start(i_start):
                        continue
                    any_activated = True
                    i_cmodel = self.control_model[i_type]
                    # inpainting workaround
                    if i_cmodel.in_channels == net.shape[1]:
                        cnet = net
                    else:
                        cnet = net[:, : i_cmodel.in_channels]
                    i_control = i_cmodel(cnet, i_hint, timesteps=timesteps, **cond_kw)
                    if self.control_scales is None:
                        i_scales = [1.0] * self.num_control_scales
                    else:
                        if not isinstance(self.control_scales[i], list):
                            raise ValueError("`control_scales` should be list of list")
                        i_scales = self.control_scales[i]  # type: ignore
                    for j, (j_c, j_scale) in enumerate(zip(i_control, i_scales)):
                        ctrl[j] += j_c * j_scale
                if not any_activated:
                    ctrl = None
            cond_kw["control"] = ctrl
            cond_kw["only_mid_control"] = self.only_mid_control
        return self.unet(net, timesteps=timesteps, **cond_kw)

    def predict_eps_from_z_and_v(
        self,
        image: Tensor,
        ts: Tensor,
        v: Tensor,
    ) -> Tensor:
        num_dim = image.ndim
        return (
            extract_to(self.sqrt_alphas_cumprod, ts, num_dim) * v
            + extract_to(self.sqrt_one_minus_alphas_cumprod, ts, num_dim) * image
        )

    def predict_start_from_z_and_v(
        self,
        image: Tensor,
        ts: Tensor,
        v: Tensor,
    ) -> Tensor:
        num_dim = image.ndim
        return (
            extract_to(self.sqrt_alphas_cumprod, ts, num_dim) * image
            - extract_to(self.sqrt_one_minus_alphas_cumprod, ts, num_dim) * v
        )

    def make_control_net(self, hint_channels: Union[int, Dict[str, int]]) -> None:
        def _make(n: int) -> ControlNet:
            # temporarily make inpainting compatible
            kw = shallow_copy_dict(self.unet_kw)
            kw["in_channels"] = 4
            dtype = list(self.unet.parameters())[0].dtype
            control_model = ControlNet(hint_channels=n, **kw)  # type: ignore
            control_model = control_model.to(get_device(self), dtype=dtype)
            return control_model

        if isinstance(hint_channels, int):
            self.control_model = _make(hint_channels)
        else:
            self.control_model = nn.ModuleDict(
                {
                    key: _make(key_channels)
                    for key, key_channels in hint_channels.items()
                }
            )

    def rename_control_net(self, old: str, new: str) -> None:
        if not isinstance(self.control_model, nn.ModuleDict):
            msg = "`rename_control_net` is only available when multiple `ControlNet` are used"
            raise ValueError(msg)
        if old not in self.control_model:
            raise ValueError(f"cannot find '{old}' in current models")
        m = self.control_model.pop(old)
        self.control_model[new] = m

    def load_control_net_with(self, name: str, d: tensor_dict_type) -> None:
        if not isinstance(self.control_model, nn.ModuleDict):
            msg = "`load_control_net_with` is only available when multiple `ControlNet` are used"
            raise ValueError(msg)
        if name not in self.control_model:
            raise ValueError(f"cannot find '{name}' in current models")
        self.control_model[name].load_state_dict(d)

    def detach_control_net(self) -> None:
        if self.control_model is not None:
            self.control_model.to("cpu")
            del self.control_model
            self.control_model = None

    # internal

    def _q_sample(
        self,
        net: Tensor,
        timesteps: Tensor,
        noise: Optional[Tensor] = None,
    ) -> Tensor:
        return self.q_sampler.q_sample(net, timesteps, noise)

    def _preprocess(self, net: Tensor, *, deterministic: bool = False) -> Tensor:
        return net

    def _get_cond(self, cond: Any) -> cond_type:
        if self.condition_model is None:
            msg = "should not call `get_cond` when `condition_model` is not provided"
            raise ValueError(msg)
        if not isinstance(cond, dict):
            return self.condition_model(cond)
        for k, v in cond.items():
            if not is_misc_key(k):
                cond[k] = self.condition_model(v)
        return cond

    def _initialize_condition_model(
        self,
        condition_model: Optional[str],
        condition_config: Optional[Dict[str, Any]],
        condition_learnable: bool,
        use_pretrained_condition: bool,
    ) -> None:
        if condition_model is None:
            self.condition_model = None
            return
        if condition_model == self.identity_condition_model:
            self.condition_model = nn.Identity()
            return
        specialized = specialized_condition_models.get(condition_model)
        if specialized is not None:
            m = safe_execute(specialized, condition_config or {})
        else:
            kwargs = condition_config or {}
            kwargs.setdefault("pretrained", use_pretrained_condition)
            m = DLZoo.load_pipeline(condition_model, **kwargs).build_model.model
        if not condition_learnable:
            freeze(m)
        self.condition_model = make_condition_model(condition_model, m)

    def _register_noise_schedule(
        self,
        timesteps: int,
        given_betas: Optional[np.ndarray],
        beta_schedule: str,
        linear_start: float,
        linear_end: float,
        cosine_s: float,
    ) -> None:
        if given_betas is not None:
            betas = given_betas
            timesteps = len(betas)
        else:
            args = beta_schedule, timesteps, linear_start, linear_end, cosine_s
            betas = make_beta_schedule(*args)

        alphas = 1.0 - betas
        alphas_cumprod = np.cumprod(alphas, axis=0)
        alphas_cumprod_prev = np.append(1.0, alphas_cumprod[:-1])

        self.t = timesteps
        self.linear_start = linear_start
        self.linear_end = linear_end

        self.register_buffer("betas", to_torch(betas))
        self.register_buffer("alphas_cumprod", to_torch(alphas_cumprod))
        self.register_buffer("alphas_cumprod_prev", to_torch(alphas_cumprod_prev))

        # cache for q(x_t | x_{t-1})
        self.register_buffer("sqrt_alphas_cumprod", to_torch(np.sqrt(alphas_cumprod)))
        one_m_cumprod = 1.0 - alphas_cumprod
        sqrt_one_m_cumprod = np.sqrt(one_m_cumprod)
        self.register_buffer(
            "sqrt_one_minus_alphas_cumprod",
            to_torch(sqrt_one_m_cumprod),
        )

        # cache for q(x_{t-1} | x_t, x_0)
        a0, a1 = self.v_posterior, 1.0 - self.v_posterior
        p0 = a0 * betas
        p1 = a1 * betas * (1.0 - alphas_cumprod_prev) / one_m_cumprod
        posterior_variance = p0 + p1
        self.register_buffer("posterior_variance", to_torch(posterior_variance))
        self.register_buffer(
            "posterior_log_variance_clipped",
            to_torch(np.log(np.maximum(posterior_variance, 1e-20))),
        )
        self.register_buffer("posterior_coef1", to_torch(1.0 / np.sqrt(alphas)))
        self.register_buffer(
            "posterior_coef2",
            to_torch((1.0 - alphas) / sqrt_one_m_cumprod),
        )

        # TODO : check these!
        if self.parameterization == "eps":
            lvlb_weights = (
                0.5
                * self.betas**2
                / (
                    self.posterior_variance
                    * to_torch(alphas)
                    * (1.0 - self.alphas_cumprod)
                )
            )
        elif self.parameterization == "x0":
            lvlb_weights = to_torch(0.25 * np.sqrt(alphas_cumprod) / one_m_cumprod)
        elif self.parameterization == "v":
            lvlb_weights = torch.ones_like(
                self.betas**2
                / (
                    2
                    * self.posterior_variance
                    * to_torch(alphas)
                    * (1.0 - self.alphas_cumprod)
                )
            )
        else:
            msg = f"unrecognized parameterization '{self.parameterization}' occurred"
            raise NotImplementedError(msg)
        lvlb_weights[0] = lvlb_weights[1]
        self.register_buffer("lvlb_weights", lvlb_weights, persistent=False)
        assert not torch.isnan(self.lvlb_weights).any()


__all__ = [
    "DDPM",
    "DDPMStep",
]



import math
import torch

from abc import abstractmethod
from abc import ABCMeta
from torch import nn
from torch import Tensor
from typing import Any
from typing import Tuple
from typing import Optional
from cftool.types import tensor_dict_type

from .constants import TEXT_KEY
from ...schema import TrainerState
from ...schema import IDLModel
from ...constants import INPUT_KEY
from ...constants import PREDICTIONS_KEY


class IPerceptor(IDLModel, metaclass=ABCMeta):
    def __init__(self, img_size: int, context_length: int):
        super().__init__()
        self.img_size = img_size
        self.context_length = context_length
        self.logit_scale = nn.Parameter(torch.tensor(math.log(1 / 0.07)))

    @abstractmethod
    def encode_image(self, image: Tensor) -> Tensor:
        pass

    @abstractmethod
    def encode_text(self, text: Tensor) -> Tensor:
        pass

    def get_forward_args(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: Optional["TrainerState"] = None,
        **kwargs: Any,
    ) -> Tuple[Any, ...]:
        return batch[INPUT_KEY], batch[TEXT_KEY]

    def forward(self, image: Tensor, text: Tensor) -> tensor_dict_type:
        image_features = self.encode_image(image)
        text_features = self.encode_text(text)
        logit_scale = self.logit_scale.exp()
        logits_per_image = logit_scale * image_features @ text_features.t()
        return {PREDICTIONS_KEY: logits_per_image}


__all__ = [
    "IPerceptor",
]



import math
import torch

import numpy as np
import torch.nn.functional as F

from abc import abstractmethod
from abc import ABCMeta
from torch import Tensor
from typing import Any
from typing import Dict
from typing import Tuple
from typing import Optional
from typing import Protocol

from .schema import ISampler
from .schema import IDiffusion
from .schema import DDPMQSampler
from .schema import UncondSamplerMixin
from ..utils import cond_type
from ..utils import get_timesteps
from ...ae.vq import AutoEncoderVQModel
from .....misc.toolkit import get_device


class IGetModelOutput(Protocol):
    def __call__(self, image: Tensor, ts: Tensor) -> Tensor:
        pass


class IGetDenoised(Protocol):
    def __call__(self, model_output: Tensor, ts: Tensor) -> Tensor:
        pass


class DDIMQSampler(DDPMQSampler):
    def reset_buffers(self, discretize: str, total_step: int) -> None:  # type: ignore
        if discretize == "uniform":
            span = self.model.t // (total_step + int(self.model.t % total_step == 0))
            q_timesteps = np.array(list(range(0, self.model.t, span)))
        elif discretize == "quad":
            end = math.sqrt(self.model.t * 0.8)
            q_timesteps = (np.linspace(0, end, total_step) ** 2).astype(int)
        else:
            raise ValueError(f"unrecognized discretize method '{discretize}' occurred")
        q_timesteps += 1
        q_timesteps = np.minimum(q_timesteps, self.model.t - 1)
        alphas = self.model.alphas_cumprod
        self.alphas = alphas[q_timesteps]
        self.timesteps = q_timesteps
        super().reset_buffers(torch.sqrt(self.alphas), torch.sqrt(1.0 - self.alphas))


class DDIMMixin(ISampler, UncondSamplerMixin, metaclass=ABCMeta):
    def __init__(
        self,
        model: IDiffusion,
        *,
        eta: float = 0.0,
        discretize: str = "uniform",
        unconditional_cond: Optional[Any] = None,
        unconditional_guidance_scale: float = 1.0,
        temperature: float = 1.0,
        noise_dropout: float = 0.0,
        quantize_denoised: bool = False,
        default_steps: int = 50,
    ):
        if model.parameterization not in ("eps", "v"):
            raise ValueError("only `v` / `eps` parameterization is supported in `ddim`")
        super().__init__(model)
        self.eta = eta
        self.discretize = discretize
        self.unconditional_cond = unconditional_cond
        self.unconditional_guidance_scale = unconditional_guidance_scale
        self.temperature = temperature
        self.noise_dropout = noise_dropout
        self.quantize_denoised = quantize_denoised
        self.default_steps = default_steps

    # abstract

    @abstractmethod
    def sample_step_core(
        self,
        image: Tensor,
        cond: Optional[cond_type],
        step: int,
        total_step: int,
        get_model_output: IGetModelOutput,
        get_denoised: IGetDenoised,
        *,
        eta: float,
        discretize: str,
        unconditional_cond: Optional[Any],
        unconditional_guidance_scale: float,
        temperature: float,
        noise_dropout: float,
        quantize_denoised: bool,
        **kwargs: Any,
    ) -> Tensor:
        pass

    # inheritance

    @property
    def q_sampler(self) -> DDIMQSampler:
        return self._q_sampler

    @property
    def sample_kwargs(self) -> Dict[str, Any]:
        return dict(
            eta=self.eta,
            discretize=self.discretize,
            unconditional_cond=self.unconditional_cond,
            unconditional_guidance_scale=self.unconditional_guidance_scale,
            temperature=self.temperature,
            noise_dropout=self.noise_dropout,
            quantize_denoised=self.quantize_denoised,
        )

    def sample_step(
        self,
        image: Tensor,
        cond: Optional[cond_type],
        step: int,
        total_step: int,
        *,
        eta: float = 0.0,
        discretize: str = "uniform",
        unconditional_cond: Optional[Any] = None,
        unconditional_guidance_scale: float = 1.0,
        temperature: float = 1.0,
        noise_dropout: float = 1.0,
        quantize_denoised: bool = False,
        **kwargs: Any,
    ) -> Tensor:
        if step == 0 and not self.initialized:
            self._reset_buffers(
                eta,
                discretize,
                total_step,
                unconditional_cond,
                unconditional_guidance_scale,
            )
        self._register_temp_buffers(image, step, total_step)
        return self.sample_step_core(
            image,
            cond,
            step,
            total_step,
            lambda img, ts: self._uncond_denoise(img, ts, cond, step, total_step),
            lambda model_output, ts: self._get_denoised_and_pred_x0(
                model_output,
                ts,
                image,
                quantize_denoised,
                temperature,
                noise_dropout,
            )[0],
            eta=eta,
            discretize=discretize,
            unconditional_cond=unconditional_cond,
            unconditional_guidance_scale=unconditional_guidance_scale,
            temperature=temperature,
            noise_dropout=noise_dropout,
            quantize_denoised=quantize_denoised,
            **kwargs,
        )

    # internal

    def _register_temp_buffers(self, image: Tensor, step: int, total_step: int) -> None:
        b = image.shape[0]
        device = image.device
        index = total_step - step - 1
        t = self.ddim_timesteps[index]
        t_next = self.ddim_timesteps[max(index - 1, 0)]
        self._ts = get_timesteps(t, b, device)
        self._ts_next = get_timesteps(t_next, b, device)
        extract = lambda base: torch.full(
            (b, 1, 1, 1),
            base[index],
            dtype=image.dtype,
            device=device,
        )
        self._at = extract(self.ddim_alphas)
        self._a_prev_t = extract(self.ddim_alphas_prev)
        self._sigmas_t = extract(self.ddim_sigmas)
        self._sqrt_one_minus_at = extract(self.ddim_sqrt_one_minus_alphas)

    def _get_denoised_and_pred_x0(
        self,
        model_output: Tensor,
        ts: Tensor,
        image: Tensor,
        quantize_denoised: bool,
        temperature: float,
        noise_dropout: float,
    ) -> Tuple[Tensor, Tensor]:
        if self.model.parameterization != "v":
            eps = model_output
        else:
            eps = self.model.predict_eps_from_z_and_v(image, ts, model_output)
        if self.model.parameterization != "v":
            pred_x0 = (image - self._sqrt_one_minus_at * eps) / self._at.sqrt()
        else:
            pred_x0 = self.model.predict_start_from_z_and_v(image, ts, model_output)
        if quantize_denoised:
            err_fmt = "only {} can use `quantize_denoised`"
            first_stage = getattr(self.model, "first_stage", None)
            if first_stage is None:
                raise ValueError(err_fmt.format("model with `first_stage` model"))
            if not isinstance(first_stage, AutoEncoderVQModel):
                raise ValueError(
                    err_fmt.format(
                        "model with `AutoEncoderVQModel` as the `first_stage` model"
                    )
                )
            pred_x0 = first_stage.codebook(pred_x0).z_q
        direction = (1.0 - self._a_prev_t - self._sigmas_t**2).sqrt() * eps
        noise = self._sigmas_t * torch.randn_like(image) * temperature
        if noise_dropout > 0:
            noise = F.dropout(noise, p=noise_dropout)
        denoised = self._a_prev_t.sqrt() * pred_x0 + direction + noise
        return denoised, pred_x0

    def _reset_buffers(
        self,
        eta: float,
        discretize: str,
        total_step: int,
        unconditional_cond: Optional[Any],
        unconditional_guidance_scale: float,
    ) -> None:
        self._q_sampler = DDIMQSampler(self.model)
        self._q_sampler.reset_buffers(discretize, total_step)
        alphas = self.model.alphas_cumprod
        self.ddim_alphas = self._q_sampler.alphas
        self.ddim_alphas_prev = torch.tensor(
            [alphas[0]] + alphas[self._q_sampler.timesteps[:-1]].tolist(),
            dtype=alphas.dtype,
            device=get_device(self.model),
        )
        self.ddim_sigmas = eta * torch.sqrt(
            (1.0 - self.ddim_alphas_prev)
            / (1.0 - self.ddim_alphas)
            * (1.0 - self.ddim_alphas / self.ddim_alphas_prev)
        )
        self.ddim_sqrt_one_minus_alphas = self._q_sampler.sqrt_one_minus_alphas
        self.ddim_timesteps = self._q_sampler.timesteps.tolist()
        # unconditional conditioning
        self._reset_uncond_buffers(unconditional_cond, unconditional_guidance_scale)
        # set flag
        self.initialized = True


@ISampler.register("ddim")
class DDIMSampler(DDIMMixin):
    def sample_step_core(
        self,
        image: Tensor,
        cond: Optional[cond_type],
        step: int,
        total_step: int,
        get_model_output: IGetModelOutput,
        get_denoised: IGetDenoised,
        *,
        eta: float,
        discretize: str,
        unconditional_cond: Optional[Any],
        unconditional_guidance_scale: float,
        temperature: float,
        noise_dropout: float,
        quantize_denoised: bool,
        **kwargs: Any,
    ) -> Tensor:
        model_output = get_model_output(image, self._ts)
        denoised = get_denoised(model_output, self._ts)
        return denoised


__all__ = [
    "DDIMSampler",
]
